import asyncio
import functools
from typing import (
    TYPE_CHECKING,
    Type,
    TypeVar,
    List,
    Callable,
    Union,

)
from collections import OrderedDict
from dataclasses import dataclass

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands import (
    M_REVISION,
    M_MODEL
)
from ._base_manager import ResourcesBaseManager
from xoa_driver.internals import exceptions


@dataclass
class ModuleData:
    """Base module data.
    """
    module_id: int
    ports_count: int
    revision: str = ""


class L23ModuleData(ModuleData):
    """L23 module data
    """
    async def get_revision(self, conn) -> None:
        """Return the module name of the L23 module.
        """
        r = await M_REVISION(conn, self.module_id).get()
        self.revision = r.revision


class L47ModuleData(ModuleData):
    """L47 module data
    """
    async def get_revision(self, conn) -> None:
        """Return the module name of the L47 module.
        """
        r = await M_MODEL(conn, self.module_id).get()
        self.revision = r.model


MT = TypeVar("MT")
MTypeObtainer = Callable[[str], Type]
ModuleDataType = Union[Type["L23ModuleData"], Type["L47ModuleData"]]


class ModulesManager(ResourcesBaseManager[MT]):
    """Module Index Manager
    """

    __slots__ = ("_conn", "__m_types_obtainer")

    def __init__(self, conn: "itf.IConnection", m_types_obtainer: MTypeObtainer) -> None:
        super().__init__()
        self._conn = conn
        self.__m_types_obtainer = m_types_obtainer

    async def fill(self, ports_count: List[int], module_type: ModuleDataType) -> None:
        """Method for create and fill in."""
        assert not self._lock, "Method <fill> can be called only once."
        identities = [
            module_type(
                module_id=slot_id,
                ports_count=p_count
            )
            for slot_id, p_count in enumerate(ports_count)
            if p_count > 0
        ]
        await asyncio.gather(*[mi.get_revision(self._conn) for mi in identities])
        self._items = OrderedDict(
            (
                idnt.module_id,
                self.__m_types_obtainer(idnt.revision)(self._conn, idnt)
            )
            for idnt in identities
        )
        if len(self) == 0:
            raise exceptions.WrongTesterError()
        coros = list(self._items.values())
        await asyncio.gather(*coros)  # type: ignore

    fill_l23 = functools.partialmethod(fill, module_type=L23ModuleData)
    fill_l47 = functools.partialmethod(fill, module_type=L47ModuleData)



#: Chassis Commands

from dataclasses import dataclass
import ipaddress
import typing
import functools

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
from .enums import *  # noqa: F403


@register_command
@dataclass
class C_LOGON:
    """
    You log on to the chassis by setting the value of this command to the correct
    password for the chassis. All other commands will fail if the session has not
    been logged on.
    """

    code: typing.ClassVar[int] = 1
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        password: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the password value.

    def set(self, password: str) -> "Token":
        """Set the password for creating a tester management session and logging on to the tester.

        :param password: password for creating a tester management session and logging on to the tester.
        :type password: str
        """
        return Token(self._connection, build_set_request(self, password=password))


@register_command
@dataclass
class C_OWNER:
    """
    Identify the owner of the management session. The name can be any short quoted
    string up to eight characters long. This name will be used when reserving ports
    prior to updating their configuration. There is no authentication of the users,
    and the chassis does not have any actual user accounts. Multiple concurrent
    connections may use the same owner name, but only one connection can have any
    particular resource reserved at any given time. Until an owner is specified the
    chassis configuration can only be read. Once specified, the session can reserve
    ports for that owner, and will inherit any existing reservations for that owner
    retained at the chassis. Maximum 32 ASCII characters.
    """

    code: typing.ClassVar[int] = 2
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        username: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the name of the owner of this session.

    @dataclass(frozen=True)
    class GetDataAttr:
        username: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the name of the owner of this session.

    def get(self) -> "Token[GetDataAttr]":
        """Get the username of this chassis management session.

        :return: The username of this chassis management session.
        :rtype: C_OWNER.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, username: str) -> "Token":
        """Set the username of this chassis management session.

        :param username: the username of this chassis management session.
        :type username: str
        """
        return Token(self._connection, build_set_request(self, username=username))


@register_command
@dataclass
class C_KEEPALIVE:
    """
    You can request this value from the chassis, simply to let it (as well as and
    any routers and proxies between you) know that the connection is still valid.
    """

    code: typing.ClassVar[int] = 3
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        tick_count: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, an increasing number from the chassis.

    def get(self) -> "Token[GetDataAttr]":
        """Get the tick count value.

        :return: an increasing number from the chassis.
        :rtype: C_KEEPALIVE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_TIMEOUT:
    """
    The maximum number of idle seconds allowed before the connection is timed out by
    the tester.
    """

    code: typing.ClassVar[int] = 4
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        second_count: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the maximum idle interval, default is 130 seconds.

    @dataclass(frozen=True)
    class GetDataAttr:
        second_count: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the maximum idle interval, default is 130 seconds.

    def get(self) -> "Token[GetDataAttr]":
        """Get the timeout value.

        :return: the maximum idle interval, default is 130 seconds.
        :rtype: C_TIMEOUT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, second_count: int) -> "Token":
        """Set the timeout value.

        :param second_count: the maximum idle interval, default is 130 seconds.
        :type second_count: int
        """
        return Token(self._connection, build_set_request(self, second_count=second_count))


@register_command
@dataclass
class C_RESERVATION:
    """
    You set this command to reserve, release, or relinquish the chassis itself.
    The chassis must be reserved before any of the chassis-level parameters can be
    changed. The owner of the session must already have been specified.
    Reservation will fail if any modules or ports are reserved for other users.

    NOTICE: Before reserve Tester need to reserve all the ports on it, otherwise
    ``<STATUS_NOTVALID>``
    """

    code: typing.ClassVar[int] = 5
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        operation: XmpField[xt.XmpByte] = XmpField(
            xt.XmpByte, choices=ReservedAction
        )  # coded byte, containing the operation to perform. The reservation parameters are asymmetric with respect to set/get. When set, it contains the operation to perform. When get, it contains the status.

    @dataclass(frozen=True)
    class GetDataAttr:
        operation: XmpField[xt.XmpByte] = XmpField(
            xt.XmpByte, choices=ReservedStatus
        )  # coded byte, containing the operation to perform. The reservation parameters are asymmetric with respect to set/get. When set, it contains the operation to perform. When get, it contains the status.

    def get(self) -> "Token[GetDataAttr]":
        """Get the chassis reservation status.

        :return: the status of chassis reservation.
        :rtype: C_RESERVATION.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, operation: ReservedAction) -> "Token":
        """Set the chassis reservation operation to be performed.

        :param operation: reservation operation to be performed.
        :type operation: ReservedAction
        """
        return Token(self._connection, build_set_request(self, operation=operation))

    set_release = functools.partialmethod(set, ReservedAction.RELEASE)
    """Release the ownership of the tester.
    """
    set_reserve = functools.partialmethod(set, ReservedAction.RESERVE)
    """Reserve the tester.
    """
    set_relinquish = functools.partialmethod(set, ReservedAction.RELINQUISH)
    """Release the ownership of the tester from another user.
    """


@register_command
@dataclass
class C_RESERVEDBY:
    """
    Identify the user who has the chassis reserved. The empty string if the chassis
    is not currently reserved.
    """

    code: typing.ClassVar[int] = 6
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        username: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the name of the current owner of the chassis.

    def get(self) -> "Token[GetDataAttr]":
        """Get the username of the current owner of the tester.

        :return: the username of the current owner of the tester.
        :rtype: C_RESERVEDBY.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_LOGOFF:
    """
    Terminates the current scripting session. Courtesy only, the chassis will also
    handle disconnection at the TCP/IP level
    """

    code: typing.ClassVar[int] = 7
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        pass

    def set(self) -> "Token":
        """Log off from the tester and close the management session.
        """
        return Token(self._connection, build_set_request(self))


@register_command
@dataclass
class C_DOWN:
    """
    Shuts down the chassis, and either restarts it in a clean state or leaves it
    powered off.
    """

    code: typing.ClassVar[int] = 8
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        magic: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, must be the special value -1480937026.
        operation: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=ChassisShutdownAction)  # coded byte, what to do after shutting chassis down.

    def set(self, operation: ChassisShutdownAction) -> "Token":
        """Shuts down the chassis, and either restarts it in a clean state or leaves it powered off.

        :param operation: what to do after shutting chassis down.
        :type operation: ChassisShutdownAction
        """
        return Token(self._connection, build_set_request(self, magic=-1480937026, operation=operation))

    set_restart = functools.partialmethod(set, ChassisShutdownAction.RESTART)
    """Shuts down the tester and then restarts it.
    """
    set_poweroff = functools.partialmethod(set, ChassisShutdownAction.POWER_OFF)
    """Shuts down the tester and leaves it powered off.
    """


@register_command
@dataclass
class C_CAPABILITIES:
    """
    A series of integer values specifying various internal limits (aka.
    capabilities) of the chassis.
    """

    code: typing.ClassVar[int] = 9
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        version: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, chassis software build number.
        max_name_len: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, max ASCII characters in chassis name.
        max_comment_len: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, max ASCII characters in chassis comment.
        max_password_len: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, max ASCII characters in chassis password.
        max_ext_rate: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, maximum rate for external traffic.
        max_session_count: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, max number of management and scripting sessions.
        max_chain_depth: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, max chain index.
        max_module_count: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, maximum number of L23 modules.
        max_protocol_count: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, max protocol segments in a packet.
        can_stream_based_arp: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, does server support stream-based ARP/NDP?
        can_sync_traffic_start: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, does server support synchronous traffic start?
        can_read_log_files: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, can clients read debug log files from server?
        can_par_module_upgrade: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, can server handle parallel module upgrades?
        can_upgrade_timekeeper: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, is server capable of upgrading the TimeKeeper application?
        can_custom_defaults: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, can server handle custom default values for XMP parameters?
        can_latency_f2f: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, can server handle first-to-first latency mode?
        max_owner_name_length: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, max number of ASCII characters in C_OWNER name
        can_read_temperatures: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, can the server read out chassis and/or CPU temperatures? (C_TEMPERATURE ?)

    def get(self) -> "Token[GetDataAttr]":
        """Get the internal limits (capabilities) of the tester.

        :return: A series of integer values specifying various internal limits
            - chassis software build number
            - max ASCII characters in chassis name
            - max ASCII characters in chassis comment
            - max ASCII characters in chassis password
            - maximum rate for external traffic
            - max number of management and scripting sessions
            - max chain index
            - maximum number of L23 modules
            - max protocol segments in a packet
            - does server support stream-based ARP/NDP?
            - does server support synchronous traffic start?
            - can clients read debug log files from server?
            - can server handle parallel module upgrades?
            - is server capable of upgrading the TimeKeeper application?
            - can server handle custom default values for XMP parameters?
            - can server handle first-to-first latency mode?
            - max number of ASCII characters in ``C_OWNER`` name
            - can the server read out chassis and/or CPU temperatures?

        :rtype: C_CAPABILITIES.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_MODEL:
    """
    Gets the specific model of this Xena chassis.
    """

    code: typing.ClassVar[int] = 10
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        model: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the Xena model designation for the chassis.

    def get(self) -> "Token[GetDataAttr]":
        """Get the model of this Xena tester.

        :return: the model of the Xena tester
        :rtype: C_MODEL.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_SERIALNO:
    """
    Gets the unique serial number of this particular Xena chassis.
    """

    code: typing.ClassVar[int] = 11
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        serial_number: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the serial number of this chassis.

    def get(self) -> "Token[GetDataAttr]":
        """Get the serial number of this Xena tester.

        :return: the serial number of the Xena tester
        :rtype: C_SERIALNO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_VERSIONNO:
    """
    Gets the major version numbers for the chassis firmware and the Xena PCI
    driver installed on the chassis.
    """

    code: typing.ClassVar[int] = 12
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        chassis_major_version: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the chassis firmware major version number.
        pci_driver_version: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the cXena PCI driver version.

    def get(self) -> "Token[GetDataAttr]":
        """Gets the major version numbers for the tester firmware and the Xena PCI driver installed on the chassis.

        :return: the firmware major version number of the tester and the PCI driver version
        :rtype: C_VERSIONNO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_PORTCOUNTS:
    """
    Gets the number of ports in each module slot of the chassis, and indirectly
    the number of slots and modules.

    .. note::

        CFP modules return the number 8 which is the maximum number of 10G ports, but the actual number of ports can be configured dynamically using the M_CFPCONFIGEXT` command.

    """

    code: typing.ClassVar[int] = 13
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        port_counts: XmpField[xt.XmpByteList] = XmpField(xt.XmpByteList)  # list of bytes, the number of ports, typically 2 or 6, or 0 for an empty slot.

    def get(self) -> "Token[GetDataAttr]":
        """Get the number of ports in each module slot of the tester, and indirectly the number of slots and modules.

        :return: the number of ports of each module slot of the tester, 0 for an empty slot.
        :rtype: C_PORTCOUNTS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_PORTERRORS:
    """
    Gets the number of errors detected across all streams on each port of each
    test module of the chassis. The counts are ordered in sequence with those of
    the module in the lowest numbered chassis slot first. Empty slots are skipped
    so that a chassis with a 6-port and a 2-port test module will return eight
    counts regardless of which slots they are in.

    .. note::

        CFP modules return eight error counts since they can be configured as up to eight 10G ports. When in 100G and 40G mode only the first one or two counts are significant.

    .. note::

        FCS errors are included, which leads to double-counting for streams detecting lost packets using the test payload mechanism.

    """

    code: typing.ClassVar[int] = 16
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        error_count: XmpField[xt.XmpLongList] = XmpField(xt.XmpLongList)  # list of long integers, the total number of errors across all streams, and including FCS errors.

    def get(self) -> "Token[GetDataAttr]":
        """Gets the number of errors detected across all streams on each port of each
        test module of the chassis. The counts are ordered in sequence with those of
        the module in the lowest numbered chassis slot first. Empty slots are skipped
        so that a chassis with a 6-port and a 2-port test module will return eight
        counts regardless of which slots they are in.

        :return: the total number of errors across all streams, and including FCS errors.
        :rtype: C_PORTERRORS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_REMOTEPORTCOUNTS:
    """
    Gets the number of ports of each remote module. A remote module is a
    relative to the xenaserver, for example, xenal47server. The first integer in
    the returned list is always 0 because it represents the xenaserver, which is
    not a remote module.
    """

    code: typing.ClassVar[int] = 17
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        port_counts: XmpField[xt.XmpByteList] = XmpField(xt.XmpByteList)  # list of bytes, the number of ports, typically 2 or 6, or 0 for an empty slot.

    def get(self) -> "Token[GetDataAttr]":
        """Gets the number of ports of each remote module. A remote module is a
        relative to the xenaserver, for example, xenal47server. The first integer in
        the returned list is always 0 because it represents the xenaserver, which is
        not a remote module.

        :return: the number of ports of each module slot of the tester, 0 for an empty slot.
        :rtype: C_REMOTEPORTCOUNTS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_BUILDSTRING:
    """
    Identify the hostname of the PC that builds the xenaserver. It uniquely
    identifies the build of a xenaserver.
    """

    code: typing.ClassVar[int] = 19
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        build_string: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, identify the hostname of the PC that builds the xenaserver

    def get(self) -> "Token[GetDataAttr]":
        """Get the build string of the xenaserver.

        :return: build string that identifies the hostname of the PC that builds the xenaserver
        :rtype: C_BUILDSTRING.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_NAME:
    """
    The name of the chassis, as it appears at various places in the user interface.
    The name is also used to distinguish the various chassis contained within a
    testbed  and in files containing the configuration for an entire test case.
    """

    code: typing.ClassVar[int] = 20
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        chassis_name: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the name of the chassis.

    @dataclass(frozen=True)
    class GetDataAttr:
        chassis_name: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the name of the chassis.

    def get(self) -> "Token[GetDataAttr]":
        """Get the name of the tester

        :return: the name of the tester
        :rtype: C_NAME.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, chassis_name: str) -> "Token":
        """Set the name of the tester

        :param chassis_name: the name of the tester
        :type chassis_name: str
        """
        return Token(self._connection, build_set_request(self, chassis_name=chassis_name))


@register_command
@dataclass
class C_COMMENT:
    """
    The description of the chassis.
    """

    code: typing.ClassVar[int] = 21
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        comment: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the description of the chassis.

    @dataclass(frozen=True)
    class GetDataAttr:
        comment: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the description of the chassis.

    def get(self) -> "Token[GetDataAttr]":
        """Get the description of the tester.

        :return: the description of the tester
        :rtype: C_COMMENT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, comment: str) -> "Token":
        """Set the description of the tester.

        :param comment: the description of the tester
        :type comment: str
        """
        return Token(self._connection, build_set_request(self, comment=comment))


@register_command
@dataclass
class C_PASSWORD:
    """
    The password of the chassis, which must be provided when logging on to the chassis.
    """

    code: typing.ClassVar[int] = 22
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        password: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the password for the chassis.

    @dataclass(frozen=True)
    class GetDataAttr:
        password: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, containing the password for the chassis.

    def get(self) -> "Token[GetDataAttr]":
        """Get the password of the tester.

        :return: the password of the tester
        :rtype: C_PASSWORD.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, password: str) -> "Token":
        """Set the password of the tester.

        :param password: the password of the tester
        :type password: str
        """
        return Token(self._connection, build_set_request(self, password=password))


@register_command
@dataclass
class C_IPADDRESS:
    """
    The network configuration parameters of the chassis management port.
    """

    code: typing.ClassVar[int] = 24
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        ipv4_address: XmpField[xt.XmpIPV4Address] = XmpField(xt.XmpIPV4Address)  # address, the static IP address of the chassis.
        subnet_mask: XmpField[xt.XmpIPV4Address] = XmpField(xt.XmpIPV4Address)  # address, the subnet mask of the local network segment.
        gateway: XmpField[xt.XmpIPV4Address] = XmpField(xt.XmpIPV4Address)  # address, the gateway of the local network segment.

    @dataclass(frozen=True)
    class GetDataAttr:
        ipv4_address: XmpField[xt.XmpIPV4Address] = XmpField(xt.XmpIPV4Address)  # address, the static IP address of the chassis.
        subnet_mask: XmpField[xt.XmpIPV4Address] = XmpField(xt.XmpIPV4Address)  # address, the subnet mask of the local network segment.
        gateway: XmpField[xt.XmpIPV4Address] = XmpField(xt.XmpIPV4Address)  # address, the gateway of the local network segment.

    def get(self) -> "Token[GetDataAttr]":
        """Get the IP configuration information of the tester.

        :return:
            - the static IP address of the chassis
            - the subnet mask of the local network segment
            - the gateway of the local network segment
        :rtype: C_IPADDRESS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(
        self,
        ipv4_address: typing.Union[str, int, ipaddress.IPv4Address],
        subnet_mask: typing.Union[str, int, ipaddress.IPv4Address],
        gateway: typing.Union[str, int, ipaddress.IPv4Address],
    ) -> "Token":
        """the IP configuration information of the tester.

        :param ipv4_address: the static IP address of the chassis
        :type ipv4_address: typing.Union[str, int, ipaddress.IPv4Address]
        :param subnet_mask: the subnet mask of the local network segment
        :type subnet_mask: typing.Union[str, int, ipaddress.IPv4Address]
        :param gateway: the gateway of the local network segment
        :type gateway: typing.Union[str, int, ipaddress.IPv4Address]
        """
        return Token(self._connection, build_set_request(self, ipv4_address=ipv4_address, subnet_mask=subnet_mask, gateway=gateway))


@register_command
@dataclass
class C_DHCP:
    """
    Controls whether the chassis will use DHCP to get the management IP address.
    """

    code: typing.ClassVar[int] = 25
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, whether DHCP is enabled or disabled.

    @dataclass(frozen=True)
    class GetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, whether DHCP is enabled or disabled.

    def get(self) -> "Token[GetDataAttr]":
        """Get whether DHCP is enabled for getting management IP.

        :return: whether DHCP is enabled.
        :rtype: C_DHCP.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, on_off: OnOff) -> "Token":
        """Set DHCP for getting management IP.

        :param on_off: whether DHCP is enabled or disabled.
        :type on_off: OnOff
        """
        return Token(self._connection, build_set_request(self, on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable DHCP for for getting management IP.
    """
    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable on to DHCP for for getting management IP.
    """


@register_command
@dataclass
class C_MACADDRESS:
    """
    Get the MAC address for the chassis management port.
    """

    code: typing.ClassVar[int] = 26
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        mac_address: XmpField[xt.XmpMacAddress] = XmpField(xt.XmpMacAddress)  # six hex bytes, indicating the MAC address

    def get(self) -> "Token[GetDataAttr]":
        """Get the MAC address for the chassis management port.

        :return: the MAC address for the chassis management port
        :rtype: C_MACADDRESS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_HOSTNAME:
    """
    Get or set the chassis hostname used when DHCP is enabled.
    """

    code: typing.ClassVar[int] = 27
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        hostname: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, hostname for chassis (default value "xena-")

    @dataclass(frozen=True)
    class GetDataAttr:
        hostname: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, hostname for chassis (default value "xena-")

    def get(self) -> "Token[GetDataAttr]":
        """Get the chassis hostname used when DHCP is enabled.

        :return: the chassis hostname
        :rtype: C_HOSTNAME.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, hostname: str) -> "Token":
        """Set the chassis hostname.

        :param hostname: the chassis hostname
        :type hostname: str
        """
        return Token(self._connection, build_set_request(self, hostname=hostname))


@register_command
@dataclass
class C_FLASH:
    """
    Make all the test port LEDs flash on and off with a 1-second interval. This is
    helpful if you have multiple chassis mounted side by side and you need to
    identify a specific one.

    NOTICE: Require Tester to be reserved before change value.
    """

    code: typing.ClassVar[int] = 28
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, determines whether to blink all test port LEDs.

    @dataclass(frozen=True)
    class GetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, determines whether to blink all test port LEDs.

    def get(self) -> "Token[GetDataAttr]":
        """Get the status of test port LEDs.

        :return: the blinking status of test port LEDs
        :rtype: C_FLASH.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, on_off: OnOff) -> "Token":
        """Set test ports LEDs blinking status.

        :param on_off: determines whether to blink all test port LEDs.
        :type on_off: OnOff
        """
        return Token(self._connection, build_set_request(self, on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable flashing test port LEDs.
    """
    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable flashing test port LEDs.
    """


@register_command
@dataclass
class C_DEBUGLOGS:
    """
    Allows to dump all the logs of a chassis.
    """

    code: typing.ClassVar[int] = 30
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        message_length: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, length of the message.

        data: XmpField[xt.XmpHexList] = XmpField(xt.XmpHexList)  # list of hex bytes, all the logs of a chassis

    def get(self) -> "Token[GetDataAttr]":
        """Get chassis logs.

        :return: length of the message and all the logs of the chassis
        :rtype: C_DEBUGLOGS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_TEMPERATURE:
    """
    Get chassis temperature readings, if supported. Unit is millidegree Celsius.
    """

    code: typing.ClassVar[int] = 31
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        mb1_temperature: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the temperature of motherboard 1. Unit is millidegree Celsius.

        mb2_temperature: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the temperature of motherboard 2. Unit is millidegree Celsius.

        cpu_temperature: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the temperature of CPU. Unit is millidegree Celsius.

    def get(self) -> "Token[GetDataAttr]":
        """Get chassis temperature readings.

        :return:
            - the temperature of motherboard 1 (millidegree Celsius)
            - the temperature of motherboard 2 (millidegree Celsius)
            - the temperature of CPU (millidegree Celsius)
        :rtype: C_TEMPERATURE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_RESTPORT:
    """
    The TCP port used by the REST API server.
    """

    code: typing.ClassVar[int] = 32
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        tcp_port: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, containing the TCP port number (default 57911)

    @dataclass(frozen=True)
    class GetDataAttr:
        tcp_port: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, containing the TCP port number (default 57911)

    def get(self) -> "Token[GetDataAttr]":
        """Get the TCP port number used by the REST API server.

        :return: the TCP port number used by the REST API server
        :rtype: xt.XmpInt
        """
        return Token(self._connection, build_get_request(self))

    def set(self, tcp_port: int) -> "Token":
        """Set the TCP port number used by the REST API server.

        :param tcp_port: the TCP port number (default 57911)
        :type tcp_port: int
        """
        return Token(self._connection, build_set_request(self, tcp_port=tcp_port))


@register_command
@dataclass
class C_RESTENABLE:
    """
    Controls whether the chassis will run REST API server or not. The command takes
    affect only after chassis reset. To start/stop REST API server use ``C_RESTCONTROL`` command.
    """

    code: typing.ClassVar[int] = 33
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, determines whether REST API server should be enabled or disabled.

    @dataclass(frozen=True)
    class GetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, determines whether REST API server should be enabled or disabled.

    def get(self) -> "Token[GetDataAttr]":
        """Get the On/Off status of the REST API server.

        :return: the status of the REST API server, whether it is enabled.
        :rtype: xt.XmpByte
        """
        return Token(self._connection, build_get_request(self))

    def set(self, on_off: OnOff) -> "Token":
        """Set the On/Off status of the REST API server.

        :param on_off: determines whether REST API server should be enabled or disabled
        :type on_off: OnOff
        """
        return Token(self._connection, build_set_request(self, on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable the REST API server.
    """
    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable the REST API server.
    """


@register_command
@dataclass
class C_RESTCONTROL:
    """
    Controls REST API server. This command should be used with extra care as it can
    affect other users using the server.
    """

    code: typing.ClassVar[int] = 34
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        operation: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=RESTControlAction)  # coded byte, what to do with the REST API server.

    def set(self, operation: RESTControlAction) -> "Token":
        """Controlling the REST API server.

        :param operation: what to do with the REST API server
        :type operation: RESTControlAction
        """
        return Token(self._connection, build_set_request(self, operation=operation))


@register_command
@dataclass
class C_RESTSTATUS:
    """
    Gets the REST API server operation status - whether it is active (running) or
    not. To get the admin status (whether the server is enabled or disabled) use
    ``C_RESTCONTROL`` command.
    """

    code: typing.ClassVar[int] = 35
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        status: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=ServiceStatus)  # coded byte, determines the REST API server running status.

    def get(self) -> "Token[GetDataAttr]":
        """Get the operation status of th REST API server.

        :return: the operation status of th REST API server
        :rtype: C_RESTSTATUS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_WATCHDOG:
    """
    If the chassis stalls for a long time, when the timer expires the chassis will
    be rebooted automatically.
    """

    code: typing.ClassVar[int] = 36
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        timer_value: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the timer value that reboots the chassis. Unit = second.

    @dataclass(frozen=True)
    class GetDataAttr:
        timer_value: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the timer value that reboots the chassis. Unit = second.

    def get(self) -> "Token[GetDataAttr]":
        """Get the time value that reboots the chassis if it stalls for a long time.

        :return: the timer value that reboots the chassis. Unit = second.
        :rtype: C_WATCHDOG.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, timer_value: int) -> "Token":
        """Set the time value that reboots the chassis if it stalls for a long time.

        :param timer_value: the timer value that reboots the chassis
        :type timer_value: int
        """
        return Token(self._connection, build_set_request(self, timer_value=timer_value))


@register_command
@dataclass
class C_INDICES:
    """
    Gets the session indices for all current sessions on the chassis.
    """

    code: typing.ClassVar[int] = 40
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        """Returned data structure.

        Attributes:

        :attr session_ids: the session indices for all current sessions on the chassis
        :type session_ids: List[int]
        """
        session_ids: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)  # list of integers, the session indices for all current sessions on the chassis.

    def get(self) -> "Token[GetDataAttr]":
        """Gets the session indices for all current sessions on the chassis.

        :return: the session indices for all current sessions on the chassis
        :rtype: C_INDICES.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_STATSESSION:
    """
    Gets information and statistics for a particular session on the chassis.
    """

    code: typing.ClassVar[int] = 41
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    session_xindex: int

    @dataclass(frozen=True)
    class GetDataAttr:
        session_type: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=ChassisSessionType)  # coded integer, which kind of session.
        ipv4_address: XmpField[xt.XmpIPV4Address] = XmpField(xt.XmpIPV4Address)  # address, client IP address.
        owner: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the name of the session owner.
        operation_count: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, number of operations done during the session.
        requested_byte_count: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, number of bytes received by the chassis.
        responded_byte_count: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, number of bytes sent by the chassis.

    def get(self) -> "Token[GetDataAttr]":
        """Gets information and statistics for a particular session on the chassis.

        :return:
            - type of session
            - client IP address
            - the name of the session owner
            - number of operations done during the session
            - number of bytes received by the chassis
            - number of bytes sent by the chassis
        :rtype: C_STATSESSION.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, indices=[self.session_xindex]))


@register_command
@dataclass
class C_TKLICFILE:
    """
    Get Xena TimeKeeper license file content.
    """

    code: typing.ClassVar[int] = 49
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        license_content: XmpField[xt.XmpByteList] = XmpField(xt.XmpByteList)  # TODO: probably wrong type

    @dataclass(frozen=True)
    class GetDataAttr:
        license_content: XmpField[xt.XmpByteList] = XmpField(xt.XmpByteList)  # TODO: probably wrong type

    def get(self) -> "Token[GetDataAttr]":
        """Get Xena TimeKeeper license file content.

        :return: Xena TimeKeeper license file content
        :rtype: C_TKLICFILE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, license_content: str) -> "Token":
        """Set Xena TimeKeeper license file content.

        :param license_content: Xena TimeKeeper license file content
        :type license_content: str
        """
        return Token(self._connection, build_set_request(self, license_content=license_content))


@register_command
@dataclass
class C_TKLICSTATE:
    """
    Get the state of the Xena TimeKeeper license file content.
    """

    code: typing.ClassVar[int] = 50
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        license_file_state: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=TimeKeeperLicenseFileState)  # coded byte, timekeeper license state.

        license_type: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=TimeKeeperLicenseType)  # coded byte, license type.

        license_errors: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList, choices=TimeKeeperLicenseError)  # coded integers, license errors.

    def get(self) -> "Token[GetDataAttr]":
        """Get the state of the Xena TimeKeeper license file content.

        :return:
            - timekeeper license state
            - license type
            - license errors
        :rtype: C_TKLICSTATE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_FILESTART:
    """
    Initiates upload of a file to the chassis. This command should be followed by
    a sequence og ``C_FILEDATA`` parameters to provide the file content, and finally a
    ``C_FILEFINISH`` to commit the new file to the chassis.
    """

    code: typing.ClassVar[int] = 51
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        file_type: XmpField[xt.XmpHex4] = XmpField(xt.XmpHex4)  # four hex bytes, little-endian integer, the file type, should be 1.
        size: XmpField[xt.XmpHex4] = XmpField(xt.XmpHex4)  # four hex bytes, little-endian integer, the number of bytes in the file.
        time: XmpField[xt.XmpHex4] = XmpField(xt.XmpHex4)  # four hex bytes, little-endian integer, the Linux date+time of the file.
        mode: XmpField[xt.XmpHex4] = XmpField(xt.XmpHex4)  # four hex bytes, little-endian integer, the Linux permissions of the file.
        checksum: XmpField[xt.XmpHex4] = XmpField(xt.XmpHex4)  # four hex bytes, little-endian integer, the checksum of the file.
        name: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, the name and location of the file, as a full path.

    def set(self, file_type: str, size: str, time: str, mode: str, checksum: str, name: str) -> "Token":
        """Initiates upload of a file to the chassis.

        :param file_type: the file type, should be 1
        :type file_type: str
        :param size: the number of bytes in the file
        :type size: str
        :param time: he Linux date+time of the file
        :type time: str
        :param mode: the Linux permissions of the file
        :type mode: str
        :param checksum: the checksum of the file
        :type checksum: str
        :param name: the name and location of the file, as a full path
        :type name: str
        """
        return Token(self._connection, build_set_request(self, file_type=file_type, size=size, time=time, mode=mode, checksum=checksum, name=name))


@register_command
@dataclass
class C_FILEDATA:
    """
    Uploads a fragment of a file to the chassis.
    """

    code: typing.ClassVar[int] = 52
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        offset: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the position within the file.
        data_bytes: XmpField[xt.XmpHexList] = XmpField(xt.XmpHexList)  # list of hex bytes, the data content of a section of the file.

    def set(self, offset: int, data_bytes: str) -> "Token":
        """Uploads a fragment of a file to the chassis.

        :param offset: the position within the file
        :type offset: int
        :param data_bytes: the data content of a section of the file
        :type data_bytes: str
        """
        return Token(self._connection, build_set_request(self, offset=offset, data_bytes=data_bytes))


@register_command
@dataclass
class C_FILEFINISH:
    """
    Completes upload of a file to the chassis. After validation it will replace any
    existing file with the same name.
    """

    code: typing.ClassVar[int] = 53
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        magic: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, must be the special value -1480937026.

    def set(self) -> "Token":
        """Completes upload of a file to the chassis. After validation it will replace any existing file with the same name.
        """
        return Token(self._connection, build_set_request(self, magic=-1480937026))


@register_command
@dataclass
class C_TRAFFIC:
    """
    Starts or stops the traffic on a number of ports on the chassis simultaneously.
    The ports are identified by pairs of integers (module port).
    """

    code: typing.ClassVar[int] = 55
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, determines whether to start or stop traffic generation.
        module_ports: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)  # list of integers, specifies ports on modules, which should stop or start generating traffic.

    def set(self, on_off: OnOff, module_ports: typing.List[int]) -> "Token":
        """Starts or stops the traffic on a number of ports on the chassis simultaneously.

        :param on_off: determines whether to start or stop traffic generation
        :type on_off: OnOff
        :param module_ports: specifies ports on modules, which should stop or start generating traffic
        :type module_ports: typing.List[int]
        """
        return Token(self._connection, build_set_request(self, on_off=on_off, module_ports=module_ports))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Stop the traffic on a number of ports on the chassis simultaneously.
    """
    set_on = functools.partialmethod(set, OnOff.ON)
    """Start the traffic on a number of ports on the chassis simultaneously.
    """


@register_command
@dataclass
class C_VERSIONNO_MINOR:
    """
    Gets the minor version number for the chassis firmware. The full version of
    the chassis firmware is thus where the number is obtained  with the ``C_VERSIONNO``
    command and the number is obtained with the ``C_VERSIONNO_MINOR`` command.
    """

    code: typing.ClassVar[int] = 56
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        chassis_minor_version: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the chassis firmware minor version number.
        reserved_1: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, reserved.
        reserved_2: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, reserved.

    def get(self) -> "Token[GetDataAttr]":
        """Get the minor version number for the chassis firmware.

        :return:
            - the minor version number for the chassis firmware
            - reserved, 0
            - reserved, 0
        :rtype: C_VERSIONNO_MINOR.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_START:
    """
    Start traffic on N ports and each port is described by (module index, port
    index).
    """

    code: typing.ClassVar[int] = 60
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        module_ports: XmpField[xt.XmpByteList] = XmpField(xt.XmpByteList)  # list of bytes, specifies ports on modules, which should stop or start generating traffic.

    def set(self, module_ports: typing.List[int]) -> "Token":
        """Start traffic on N ports and each port is described by (module index, port index).

        :param module_ports: specifies ports on modules, which should stop or start generating traffic
        :type module_ports: typing.List[int]
        """
        return Token(self._connection, build_set_request(self, module_ports=module_ports))


@register_command
@dataclass
class C_STOP:
    """
    Stop traffic on N ports and each port is described by (module index, port index)
    """

    code: typing.ClassVar[int] = 61
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        module_ports: XmpField[xt.XmpByteList] = XmpField(xt.XmpByteList)  # list of bytes, specifies ports on modules, which should stop or start generating traffic.

    def set(self, module_ports: typing.List[int]) -> "Token":
        """Stop traffic on N ports and each port is described by (module index, port index).

        :param module_ports: specifies ports on modules, which should stop or start generating traffic
        :type module_ports: typing.List[int]
        """
        return Token(self._connection, build_set_request(self, module_ports=module_ports))


@register_command
@dataclass
class C_MULTIUSER:
    """
    Enable or disable the ability to control one resource from several different TCP
    connections.
    """

    code: typing.ClassVar[int] = 62
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(
            xt.XmpByte, choices=OnOff
        )  # coded byte, enable or disable the ability to control one resource from several different TCP connections

    @dataclass(frozen=True)
    class GetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(
            xt.XmpByte, choices=OnOff
        )  # coded byte, enable or disable the ability to control one resource from several different TCP connections

    def get(self) -> "Token[GetDataAttr]":
        """Get the status of the ability to control one resource from several different TCP connections.

        :return: the status of the ability to control one resource from several different TCP connections
        :rtype: C_MULTIUSER.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, on_off: OnOff) -> "Token":
        """Enable or disable the ability to control one resource from several different TCP connections.

        :param on_off: enable or disable the ability to control one resource from several different TCP connections
        :type on_off: OnOff
        """
        return Token(self._connection, build_set_request(self, on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable the ability to control one resource from several different TCP.
    """
    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable the ability to control one resource from several different TCP.
    """


@register_command
@dataclass
class C_SCRIPT:
    """
    To load and save CLI commands e.g. port configuration, through the binary XMP session.
    """

    code: typing.ClassVar[int] = 64
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        command_string: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, text CLI command

    def set(self, command_string: str) -> "Token":
        """Set the CLI commands through a binary XMP session.

        :param command_string: text CLI command
        :type command_string: str
        """
        return Token(self._connection, build_set_request(self, command_string=command_string))


@register_command
@dataclass
class C_TKSTATUS:
    """
    Report TimeKeeper version and status.
    """

    code: typing.ClassVar[int] = 65
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        status_string: XmpField[xt.XmpStr] = XmpField(
            xt.XmpStr
        )  # string. Version, TimeKeeper license expiration, and TimeKeeper status. The string is formatted as shown in the example below. Each line is separated by \n.

        # TimeKeeper Status\n
        # ================================================================================\n
        # TimeKeeper version 8.0.3\n
        # License expires in 33 days (including grace period)\n
        # TimeKeeper is not running\n

    def get(self) -> "Token[GetDataAttr]":
        """Get the version and status of TimeKeeper

        :return:
            Version, TimeKeeper license expiration, and TimeKeeper status. The string is formatted as shown in the example below.

            The format is shown below.

            TimeKeeper Status
            TimeKeeper version 8.0.3
            License expires in 33 days (including grace period)
            TimeKeeper is not running

        :rtype: C_TKSTATUS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_TKSVCSTATE:
    """
    Get and control TimeKeeper service state.
    """

    code: typing.ClassVar[int] = 66
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        state: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=TimeKeeperServiceAction)  # coded byte, TimeKeeper service state

    @dataclass(frozen=True)
    class GetDataAttr:
        state: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=TimeKeeperServiceStatus)  # coded byte, TimeKeeper service state

    def get(self) -> "Token[GetDataAttr]":
        """Get TimeKeeper service state

        :return: TimeKeeper service state
        :rtype: C_TKSVCSTATE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, state: TimeKeeperServiceAction) -> "Token":
        """Control TimeKeeper service state

        :param state: TimeKeeper service state
        :type state: TimeKeeperServiceAction
        """
        return Token(self._connection, build_set_request(self, state=state))

    set_stop = functools.partialmethod(set, TimeKeeperServiceAction.STOP)
    """Stop the TimerKeeper service.
    """
    set_start = functools.partialmethod(set, TimeKeeperServiceAction.START)
    """Start the TimerKeeper service.
    """
    set_restart = functools.partialmethod(set, TimeKeeperServiceAction.RESTART)
    """Restart the TimerKeeper service.
    """


@register_command
@dataclass
class C_TKCONFIG:
    """
    TimeKeeper config file content.
    """

    code: typing.ClassVar[int] = 67
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        config_file: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, TimeKeeper config file content

    @dataclass(frozen=True)
    class GetDataAttr:
        config_file: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, TimeKeeper config file content

    def get(self) -> "Token[GetDataAttr]":
        """Get TimeKeeper config file content.

        :return: TimeKeeper config file content
        :rtype: C_TKCONFIG.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, config_file: str) -> "Token":
        """Set TimeKeeper config file content.

        :param config_file: TimeKeeper config file content
        :type config_file: str
        """
        return Token(self._connection, build_set_request(self, config_file=config_file))


@register_command
@dataclass
class C_TKGPSSTATE:
    """
    Get TimeKeeper GPS status.
    """

    code: typing.ClassVar[int] = 68
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        status: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, TimeKeeper GPS status

    def get(self) -> "Token[GetDataAttr]":
        """Get TimeKeeper GPS status.

        :return: TimeKeeper GPS status
        :rtype: C_TKGPSSTATE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_TIME:
    """
    Get local chassis time in seconds.
    """

    code: typing.ClassVar[int] = 69
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        local_time: XmpField[xt.XmpLong] = XmpField(xt.XmpLong)  # long integer, local chassis time in seconds

    def get(self) -> "Token[GetDataAttr]":
        """Get local chassis time in seconds.

        :return: local chassis time in seconds
        :rtype: C_TIME.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))


@register_command
@dataclass
class C_TRAFFICSYNC:
    """
    Works just as the ``C_TRAFFIC`` command described above with an additional option to
    specify  a point in time where traffic should be started. This can be used to
    start traffic simultaneously on multiple chassis. The ports are identified by
    pairs of integers (module port).

    .. note::

        This requires that the chassis in question all use the TimeKeeper option to keep their CPU clocks synchronized.

    """

    code: typing.ClassVar[int] = 70
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class SetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, determines whether to start or stop traffic generation.
        timestamp: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long integer, the time where traffic should be started, expressed as the number of seconds since January 1 2010, 00
        module_ports: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)  # list of integers, specifies ports on modules, which should stop or start traffic generation.

    @dataclass(frozen=True)
    class GetDataAttr:
        on_off: XmpField[xt.XmpByte] = XmpField(xt.XmpByte, choices=OnOff)  # coded byte, status traffic generation.
        timestamp: XmpField[xt.XmpLong] = XmpField(
            xt.XmpLong
        )  # long integer, the time where traffic should be started, expressed as the number of seconds since January 1 2010, 00
        module_ports: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)  # list of integers, specifies ports on modules, which should stop or start traffic generation.

    def get(self) -> "Token[GetDataAttr]":
        """Get the status of traffic generation.

        :return:
            - status traffic generation
            - the time where traffic should be started, expressed as the number of seconds since January 1 2010, 00
            - ports on modules, which should stop or start traffic generation
        :rtype: C_TRAFFICSYNC.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))

    def set(self, on_off: OnOff, timestamp: int, module_ports: typing.List[int]) -> "Token":
        """Set the status of traffic generation.

        :param on_off: determines whether to start or stop traffic generation
        :type on_off: OnOff
        :param timestamp: the time where traffic should be started, expressed as the number of seconds since January 1 2010, 00
        :type timestamp: int
        :param module_ports: specifies ports on modules, which should stop or start traffic generation.
        :type module_ports: typing.List[int]
        """
        return Token(self._connection, build_set_request(self, on_off=on_off, timestamp=timestamp, module_ports=module_ports))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Stop traffic generation on the given ports simultaneously on different chassis.
    """
    set_on = functools.partialmethod(set, OnOff.ON)
    """Start traffic generation on the given ports simultaneously on different chassis.
    """


@register_command
@dataclass
class C_TKSTATUSEXT:
    """
    Report TimeKeeper version and status (extended version).
    """

    code: typing.ClassVar[int] = 71
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"

    @dataclass(frozen=True)
    class GetDataAttr:
        status_string: XmpField[xt.XmpStr] = XmpField(xt.XmpStr)  # string, extended status in JSON format. The string is formatted as shown in the example below.

        # {
        #     "FormatVersion": 1,
        #     "ApplicationVersion": 452.0,
        #     "TimeKeeperStatus":
        #     {
        #         "systemtimingstatus": "Waiting for good time source",
        #         "syncsource": "NTP",
        #         "sourcestate": "NTP server 10.0.0.110",
        #         "sourceaccuracy": " No updates yet",
        #         "versioninfo": "8.0.3",
        #         "timesincestart": "0 day(s) 0 hours 1 minutes",
        #         "timesinceboot": "0 day(s) 0 hours 2 minutes",
        #         "updatetime": 1637916837
        #     }
        # }

    def get(self) -> "Token[GetDataAttr]":
        """Get the TimeKeeper version and status.

        :return: extended status in JSON format. The string is formatted as shown in the example below.

            .. code-block::json
            
                {
                    "FormatVersion": 1,
                    "ApplicationVersion": 452.0,
                    "TimeKeeperStatus": {
                        "systemtimingstatus": "Waiting for good time source",
                        "syncsource": "NTP",
                        "sourcestate": "NTP server 10.0.0.110",
                        "versioninfo": "8.0.3",
                        "timesincestart": "0 day(s) 0 hours 1 minutes",
                        "timesinceboot": "0 day(s) 0 hours 2 minutes",
                        "updatetime": 1637916837
                    }
                }
            

        :rtype: C_TKSTATUSEXT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self))



#: All available tester types.
import sys

if "xoa_driver.v2" in sys.modules:
    raise ImportError("\33[31mOnly Single interface version is allowed to being use at the same time.\33[0m")

from .internals.hli_v1.testers.l23_tester import L23Tester
from .internals.hli_v1.testers.l23ve_tester import L23VeTester
from .internals.hli_v1.testers.l47_tester import L47Tester
from .internals.hli_v1.testers.l47ve_tester import L47VeTester

import typing

GenericAnyTester = typing.Union[
    "L23Tester",
    "L23VeTester",
    "L47Tester",
    "L47VeTester",
]


__all__ = (
    "L23Tester",
    "L23VeTester",
    "L47Tester",
    "L47VeTester",
    "GenericAnyTester",
)



import typing

if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf


class ModuleInitData(typing.Protocol):
    module_id: int
    ports_count: int
    revision: str


class IPort(typing.Protocol):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        ...

    # def __await__(self):
    #     ...

    # async def _setup(self):
    #     ...



#: L23 Port Transceiver Commands

from dataclasses import dataclass
import typing

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
# from .enums import *  # noqa: F403


@register_command
@dataclass
class PX_RW:
    """
    Provides read and write access to the register interface supported by the port transceiver.  It
    is possible to both read and write register values.
    """

    code: typing.ClassVar[int] = 501
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _page_xindex: int
    _register_xaddress: int

    @dataclass(frozen=True)
    class SetDataAttr:
        value: XmpField[xt.XmpHex4] = XmpField(xt.XmpHex4)  # 4 hex bytes, register value of the port transceiver

    @dataclass(frozen=True)
    class GetDataAttr:
        value: XmpField[xt.XmpHex4] = XmpField(xt.XmpHex4)  # 4 hex bytes, register value of the port transceiver

    def get(self) -> "Token[GetDataAttr]":
        """Get the register value of a transceiver.

        :return: the register value of a transceiver
        :rtype: PX_RW.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._page_xindex, self._register_xaddress]))

    def set(self, value: str) -> "Token":
        """Set the register value of a transceiver.

        :param value: register value of a transceiver
        :type value: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._page_xindex, self._register_xaddress], value=value))


@register_command
@dataclass
class PX_RW_SEQ:
    """
    .. versionadded: v1.1

    :term:`I2C<I2C>` sequential access to a transceiver's register. When invoked, the ``<byte_count>`` number of bytes will be read or written in one I2C transaction, in which the ``<value>`` is read or written with only a single register address setup. A subsequent invocation will perform a second I2C transaction in the same manner. ``<_page_xindex>``: the transceiver page address, integer, 0-255. ``<_register_xaddress>``: the address within the page, integer, 0-255.
    """

    code: typing.ClassVar[int] = 503
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _page_xindex: int
    _register_xaddress: int
    _byte_count: int

    @dataclass(frozen=True)
    class SetDataAttr:
        value: XmpField[xt.XmpHexList] = XmpField(xt.XmpHexList)  # the bytes to be read or written in one I2C transaction. The number of bytes in the ``<value>`` equals ``<byte_count>``.

    @dataclass(frozen=True)
    class GetDataAttr:
        value: XmpField[xt.XmpHexList] = XmpField(xt.XmpHexList)  # the bytes to be read or written in one I2C transaction. The number of bytes in the ``<value>`` equals ``<byte_count>``.

    def get(self) -> "Token[GetDataAttr]":
        """Get the register value of a transceiver in one I2C transaction.

        :return: the register value of a transceiver
        :rtype: PX_RW_SEQ.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._page_xindex, self._register_xaddress, self._byte_count]))

    def set(self, value: str) -> "Token":
        """Set the register value of a transceiver in one I2C transaction.

        :param value: register value of a transceiver
        :type value: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._page_xindex, self._register_xaddress, self._byte_count], value=value))


@register_command
@dataclass
class PX_MII:
    """Provides access to the register interface supported by the media-independent interface (MII) transceiver.  It
    is possible to both read and write register values."""

    code: typing.ClassVar[int] = 537
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _register_xaddress: int

    @dataclass(frozen=True)
    class SetDataAttr:
        value: XmpField[xt.XmpHex2] = XmpField(xt.XmpHex2)  # 2 hex bytes, register value of the transceiver

    @dataclass(frozen=True)
    class GetDataAttr:
        value: XmpField[xt.XmpHex2] = XmpField(xt.XmpHex2)  # 2 hex bytes, register value of the transceiver

    def get(self) -> "Token[GetDataAttr]":
        """Get the register value of a transceiver.

        :return: the register value of a transceiver
        :rtype: PX_MII.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._register_xaddress]))

    def set(self, value: str) -> "Token":
        """Set the register value of a transceiver.

        :param value: register value of a transceiver
        :type value: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._register_xaddress], value=value))


@register_command
@dataclass
class PX_TEMPERATURE:
    """
    Transceiver temperature in degrees Celsius.
    """

    code: typing.ClassVar[int] = 538
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    @dataclass(frozen=True)
    class GetDataAttr:
        integral_part: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)  # byte, temperature value before the decimal digit.
        fractional_part: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)  # byte, 1/256th of a degree Celsius after the decimal digit.

    def get(self) -> "Token[GetDataAttr]":
        """Get transceiver temperature in degrees Celsius.

        :return: temperature value before the decimal digit, and 1/256th of a degree Celsius after the decimal digit.
        :rtype: PX_TEMPERATURE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))



import asyncio
from typing import (
    Dict,
    List,
    Callable,
    Final,
    Any,
    Coroutine,
)
from collections import defaultdict

ON_EVT_DISCONNECTED: Final[int] = -1

CB = Callable[..., Coroutine[Any, None, None]]

#TODO: add type annotation for callback function


class EventsObserver:

    __slots__ = ("__events", "__loop", )

    def __init__(self) -> None:
        self.__events: Dict[int, List[CB]] = defaultdict(list)
        self.__loop = asyncio.get_event_loop()

    def dispatch(self, evt: int, *args, **kwargs) -> None:
        for evt_func in self.__events.get(evt, []):
            self.__loop.create_task(
                evt_func(*args, **kwargs)
            ).add_done_callback(self.__handle_exceptions)

    def __handle_exceptions(self, fut: "asyncio.Future") -> None:
        if e := fut.exception():
            raise e

    def subscribe(self, evt: int, func: CB) -> None:
        self.__events[evt].append(func)



from typing import (
    TYPE_CHECKING,
    List,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands import (
    P_CAPTURE,
    PC_TRIGGER,
    PC_KEEP,
    PC_STATS,
    PC_EXTRA,
    PC_PACKET,
)


class ObtainCaptured:
    """Obtain info of captured packets.
    """

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, capture_pkt_idx: int) -> None:
        self.extra = PC_EXTRA(conn, module_id, port_id, capture_pkt_idx)
        """Obtains extra information about a captured packet on a L23 port.
        Representation of PC_EXTRA
        """

        self.packet = PC_PACKET(conn, module_id, port_id, capture_pkt_idx)
        """Obtains raw bytes of a captured packet on a L23 port.
        Representation of PC_PACKET
        """


class PortCapture:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.__conn = conn
        self.__module_id = module_id
        self.__port_id = port_id
        self.state = P_CAPTURE(conn, module_id, port_id)
        """L23 port traffic capture.
        Representation of P_CAPTURE
        """

        self.trigger = PC_TRIGGER(conn, module_id, port_id)
        """L23 port packet capture triggering criteria.
        Representation of PC_TRIGGER
        """

        self.keep = PC_KEEP(conn, module_id, port_id)
        """Keeping captured packets on a L23 port.
        Representation of PC_KEEP
        """

        self.stats = PC_STATS(conn, module_id, port_id)
        """L23 port's number of packets in the capture buffer.
        Representation of PC_STATS
        """

    async def obtain_captured(self) -> List[ObtainCaptured]:
        # TODO: check better title for this operation
        stats = await self.stats.get()
        return [
            ObtainCaptured(self.__conn, self.__module_id, self.__port_id, idx)
            for idx in range(stats.packets)
        ]



#: All exception classes which can be propagated to the upper level.

from .internals.exceptions import (
    WrongModuleError,
    WrongTesterError,
    WrongTesterPasswordError,
)
from .internals.core.transporter.exceptions import (
    EstablishConnectionError,
    BadStatus,
)

__all__ = (
    "WrongModuleError",
    "WrongTesterError",
    "WrongTesterPasswordError",
    "EstablishConnectionError",
    "BadStatus",
)



#: L23 Port Length Term Commands

from dataclasses import dataclass
import typing
import functools

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
from .enums import *  # noqa: F403


@register_command
@dataclass
class PL_INDICES:
    """
    The full list of which length terms are defined for a port. These are the sub-
    index values that are used for the parameters defining the length-based matching
    of packets received for the port. Setting the value of this parameter creates a
    new empty length term for each value that is not already in use, and deletes
    each length term that is not mentioned in the list. The same can be accomplished
    one- length-term-at-a-time using the PL_CREATE and PL_DELETE commands.
    """

    code: typing.ClassVar[int] = 207
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    @dataclass(frozen=True)
    class SetDataAttr:
        length_term_xindices: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)  # list of integers, the sub-index of a length term definition for the port.

    @dataclass(frozen=True)
    class GetDataAttr:
        length_term_xindices: XmpField[xt.XmpIntList] = XmpField(xt.XmpIntList)  # list of integers, the sub-index of a length term definition for the port.

    def get(self) -> "Token[GetDataAttr]":
        """Get the full list of which length terms are defined for a port.

        :return: the full list of which length terms are defined for a port
        :rtype: PL_INDICES.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, length_term_xindices: typing.List[int]) -> "Token":
        """Create a new empty length term for each value that is not already in use, and deletes each length term that is not mentioned in the list.

        :param length_term_xindices: the list of indices of length terms to be created on a port.
        :type length_term_xindices: List[int]
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, length_term_xindices=length_term_xindices))


@register_command
@dataclass
class PL_CREATE:
    """
    Creates an empty length term definition with the specified sub-index value.
    """

    code: typing.ClassVar[int] = 208
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _length_term_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        pass

    def set(self) -> "Token":
        """Creates an empty length term definition with the specified sub-index value.
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._length_term_xindex],
            ),
        )


@register_command
@dataclass
class PL_DELETE:
    """
    Deletes the length term definition with the specified sub-index value. A length
    term cannot be deleted while it is used in the condition of any filter for the
    port.
    """

    code: typing.ClassVar[int] = 209
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _length_term_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        pass

    def set(self) -> "Token":
        """Deletes the length term definition with the specified sub-index value.
        """
        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._length_term_xindex],
            ),
        )


@register_command
@dataclass
class PL_LENGTH:
    """
    The specification for a length-based check that is applied on the packets
    received on the port.
    """

    code: typing.ClassVar[int] = 210
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _length_term_xindex: int

    @dataclass(frozen=True)
    class SetDataAttr:
        length_check_type: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=LengthCheckType)  # coded integer, whether to test for shorter-than or longer-than.
        size: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the value to compare the packet length against.

    @dataclass(frozen=True)
    class GetDataAttr:
        length_check_type: XmpField[xt.XmpInt] = XmpField(xt.XmpInt, choices=LengthCheckType)  # coded integer, whether to test for shorter-than or longer-than.
        size: XmpField[xt.XmpInt] = XmpField(xt.XmpInt)  # integer, the value to compare the packet length against.

    def get(self) -> "Token[GetDataAttr]":
        """Get the configuration of a length-based check that is applied on the packets received on a port.

        :return: whether to test for shorter-than or longer-than, and the value to compare the packet length against
        :rtype: PL_LENGTH.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._length_term_xindex]))

    def set(self, length_check_type: LengthCheckType, size: int) -> "Token":
        """Set the configuration of a length-based check that is applied on the packets received on a port.

        :param length_check_type: whether to test for shorter-than or longer-than
        :type length_check_type: LengthCheckType
        :param size: the value to compare the packet length against
        :type size: int
        """
        return Token(
            self._connection,
            build_set_request(self, module=self._module, port=self._port, indices=[self._length_term_xindex], length_check_type=length_check_type, size=size),
        )

    set_at_most = functools.partialmethod(set, LengthCheckType.AT_MOST)
    """Set the length check to be short than or equal to the given length (at most).
    """
    set_at_least = functools.partialmethod(set, LengthCheckType.AT_LEAST)
    """Set the length check to be longer than or equal to the given length (at least).
    """



from typing import (
    List,
    Type,
    Union,
    Generic,
    TypeVar,
    TYPE_CHECKING,
)
import dataclasses
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.core.commands import (
        PS_MODIFIERCOUNT,
        PS_MODIFIEREXTCOUNT
    )
    CT = Union[Type[PS_MODIFIERCOUNT], Type[PS_MODIFIEREXTCOUNT]]

MT = TypeVar("MT")


class ModifiersManager(Generic[MT]):
    def __init__(self, conn: "itf.IConnection", kind, count_type: "CT", modifier_type: Type[MT]) -> None:
        self.__conn = conn
        self.__kind = kind
        self.__modifier_type = modifier_type

        self._count = count_type(self.__conn, *kind)
        self.__items: List[MT] = []

    async def _populate(self) -> None:
        count = dataclasses.astuple(await self._count.get())[0]  # modifier_count or ext_modifier_count
        self.__items = [
            self.__modifier_type(
                self.__conn,
                *self.__kind,
                idx
            )
            for idx in range(count)
        ]

    async def configure(self, number: int) -> None:
        await self._count.set(number)
        await self._populate()

    async def clear(self) -> None:
        await self._count.set(0)
        self.__items.clear()

    def __len__(self) -> int:
        return len(self.__items)

    def __iter__(self):
        return iter(self.__items)

    def obtain(self, idx: int) -> MT:
        """Access modifier of the index value"""
        return self.__items[idx]



#: Low-Level interface


from xoa_driver.internals.core import commands
from xoa_driver.internals.core.registry import COMMANDS_REGISTRY
from xoa_driver.internals.core.transporter import TransportationHandler
from xoa_driver.internals.core.transporter.funcs import establish_connection


__all__ = (
    "commands",
    "COMMANDS_REGISTRY",
    "TransportationHandler",
    "establish_connection"
)



#: All available test port types.
import sys

if "xoa_driver.v2" in sys.modules:
    raise ImportError("\33[31mOnly Single interface version is allowed to being use at the same time.\33[0m")

import typing

from .internals.hli_v1.ports.port_l47.main import PortL47

from .internals.hli_v1.ports.port_l23.port_l23ve import PortL23VE
from .internals.hli_v1.ports.port_l23.chimera.port_chimera import (
    PortChimera,
    PChi100G5S2P,
    PChi100G5S2P_b,
    PChi40G2S2P,
)

from .internals.hli_v1.ports.port_l23.bases.port_l23 import BasePortL23
from xoa_driver.internals.hli_v1.ports.port_l23.family_combi import (
    POdin1G4S4PCombi,
    POdin1G4S4PCombi_b,
    POdin10G4S2PCombi,
    POdin10G4S2PCombi_b,
)

from xoa_driver.internals.hli_v1.ports.port_l23.family_d import (
    POdin1G3S2PT,
    POdin1G3S6P,
    POdin1G3S6P_b,
    POdin1G3S6PE,
)

from xoa_driver.internals.hli_v1.ports.port_l23.family_e import (
    POdin5G4S6PCU,
    POdin10G3S2PCU,
    POdin10G3S6PCU,
    POdin10G5S6PCU,
    POdin10G5S6PCU_b,
)

from xoa_driver.internals.hli_v1.ports.port_l23.family_f import (
    POdin10G1S2P,
    POdin10G1S2P_b,
    POdin10G1S2P_c,
    POdin10G1S2P_d,
    POdin10G1S2PT,
    POdin10G1S6P,
    POdin10G1S6P_b,
    POdin10G1S12P,
    POdin40G2S2P,
    POdin40G2S2PB,
)

from xoa_driver.internals.hli_v1.ports.port_l23.family_g import (
    PLoki100G3S1P,
    PLoki100G3S1P_b,
    PLoki100G3S1PSE,
    PLoki100G3S1PB,
)

from xoa_driver.internals.hli_v1.ports.port_l23.family_h import (
    PLoki100G5S1P,
    POdin100G3S1P,
)

from xoa_driver.internals.hli_v1.ports.port_l23.family_i import PLoki100G5S2P
from xoa_driver.internals.hli_v1.ports.port_l23.family_j import PThor100G5S4P
from xoa_driver.internals.hli_v1.ports.port_l23.family_k import PThor400G7S1P
from xoa_driver.internals.hli_v1.ports.port_l23.family_l import (
    PThor400G7S1P_b,
    PThor400G7S1P_c,
    PThor400G7S1P_d,
)

from xoa_driver.internals.hli_v1.ports.port_l23.family_l1 import (
    PFreya800G1S1P_a,
    PFreya800G1S1P_b,
    PFreya800G1S1POSFP_a,
    PFreya800G4S1P_a,
    PFreya800G4S1POSFP_a,
    PFreya800G4S1P_b,
)

from xoa_driver.internals.hli_v1.ports.port_l23.family_m import POdin1G3S6PT1RJ45

GenericL23Port = typing.Union[
    "POdin1G3S6P",
    "POdin1G3S6P_b",
    "POdin1G3S6PE",
    "POdin1G3S2PT",
    "POdin5G4S6PCU",
    "POdin10G5S6PCU",
    "POdin10G5S6PCU_b",
    "POdin10G3S6PCU",
    "POdin10G3S2PCU",
    "POdin10G1S2P",
    "POdin10G1S2P_b",
    "POdin10G1S2P_c",
    "POdin10G1S6P",
    "POdin10G1S6P_b",
    "POdin10G1S2PT",
    "POdin10G1S2P_d",
    "POdin10G1S12P",
    "POdin40G2S2P",
    "POdin40G2S2PB",
    "POdin100G3S1P",
    "PLoki100G3S1P",
    "PLoki100G3S1P_b",
    "PLoki100G3S1PSE",
    "PLoki100G3S1PB",
    "PLoki100G5S1P",
    "PLoki100G5S2P",
    "PThor100G5S4P",
    "PThor400G7S1P",
    "PThor400G7S1P_b",
    "PThor400G7S1P_c",
    "PThor400G7S1P_d",
    "PFreya800G1S1P_a",
    "PFreya800G1S1P_b",
    "PFreya800G1S1POSFP_a",
    "PFreya800G4S1P_a",
    "PFreya800G4S1P_b",
    "PFreya800G4S1POSFP_a",
    "POdin1G3S6PT1RJ45",
    "POdin1G4S4PCombi",
    "POdin1G4S4PCombi_b",
    "POdin10G4S2PCombi",
    "POdin10G4S2PCombi_b",
]

GenericAnyPort = typing.Union[
    GenericL23Port,
    "PortL47",
    "PortL23VE",
    "PortChimera",
    "PChi100G5S2P",
    "PChi100G5S2P_b",
    "PChi40G2S2P",
]

__all__ = (
    "BasePortL23",
    "POdin1G3S6P",
    "POdin1G3S6P_b",
    "POdin1G3S6PE",
    "POdin1G3S2PT",
    "POdin5G4S6PCU",
    "POdin10G5S6PCU",
    "POdin10G5S6PCU_b",
    "POdin10G3S6PCU",
    "POdin10G3S2PCU",
    "POdin10G1S2P",
    "POdin10G1S2P_b",
    "POdin10G1S2P_c",
    "POdin10G1S6P",
    "POdin10G1S6P_b",
    "POdin10G1S2PT",
    "POdin10G1S2P_d",
    "POdin10G1S12P",
    "POdin40G2S2P",
    "POdin40G2S2PB",
    "POdin100G3S1P",
    "PLoki100G3S1P",
    "PLoki100G3S1P_b",
    "PLoki100G3S1PSE",
    "PLoki100G3S1PB",
    "PLoki100G5S1P",
    "PLoki100G5S2P",
    "PThor100G5S4P",
    "PThor400G7S1P",
    "PThor400G7S1P_b",
    "PThor400G7S1P_c",
    "PThor400G7S1P_d",
    "PFreya800G1S1P_a",
    "PFreya800G1S1P_b",
    "PFreya800G1S1POSFP_a",
    "PFreya800G4S1P_a",
    "PFreya800G4S1P_b",
    "PFreya800G4S1POSFP_a",
    "POdin1G3S6PT1RJ45",
    "POdin1G4S4PCombi",
    "POdin1G4S4PCombi_b",
    "POdin10G4S2PCombi",
    "POdin10G4S2PCombi_b",
    "PortL47",
    "PortL23VE",
    "PortChimera",
    "PChi100G5S2P",
    "PChi100G5S2P_b",
    "PChi40G2S2P",
    "GenericL23Port",
    "GenericAnyPort",
)



import asyncio
from asyncio.events import AbstractEventLoop
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    AsyncGenerator,
    Optional,
    Any,
    Dict,
    List,
)
if TYPE_CHECKING:
    from .transportation_handler import TransportationHandler
    from .. import interfaces

from .token import Token
from . import exceptions


async def establish_connection(transporter: "TransportationHandler", host: str, port: int = 22606, loop: Optional["AbstractEventLoop"] = None) -> None:
    """
    Establish connection to provided host and port and assign ``<TransportationHandler>`` to it.
    """

    assert isinstance(loop, AbstractEventLoop) or loop is None, "<loop> must be an instance of AbstractEventLoop or None"
    __loop = loop if loop else asyncio.get_event_loop()
    try:
        await __loop.create_connection(lambda: transporter, host=host, port=port)
    except OSError:
        raise exceptions.EstablishConnectionError(host, port) from None


async def apply_iter(*cmd_tokens: "Token") -> AsyncGenerator[Any, None]:
    """
    Main interface for chunking the commands which need to be send to one or multiple testers at the same time.
    """

    aggregator: Dict["interfaces.IConnection", bytearray] = defaultdict(bytearray)
    queue: asyncio.Queue[asyncio.Future] = asyncio.Queue()
    for t in cmd_tokens:
        (data, fut) = await t.connection.prepare_data(t.request)
        aggregator[t.connection].extend(data)
        queue.put_nowait(fut)
    [c.send(r) for c, r in aggregator.items()]
    aggregator.clear()
    while not queue.empty():
        future = await queue.get()
        yield await future
        queue.task_done()
    await queue.join()


async def apply(*cmd_tokens: "Token") -> List[Any]:
    """
    Main interface for chunking the commands which need to be send to one or multiple testers at the same time.
    """
    assert len(cmd_tokens) <= 200, "Number of the commands is bigger then 200 for one aggregation, please use function <apply_iter> instead"
    return [f async for f in apply_iter(*cmd_tokens)]



from typing import Type
from xoa_driver.internals.core.commands import (
    C_PORTCOUNTS,
    C_MULTIUSER,
    C_TRAFFIC,
    C_TRAFFICSYNC,
    C_VERSIONNO_MINOR,
)
from xoa_driver.internals.utils.modules_manager import ModulesManager
from ._base_tester import BaseTester


from xoa_driver import modules
from xoa_driver.internals.state_storage import testers_state


TypeL23Manager = ModulesManager["modules.ModuleL23VE"]


def get_module_type(revision: str) -> Type:
    return modules.ModuleL23VE


class L23VeTester(BaseTester["testers_state.TesterLocalState"]):
    """
    This is a conceptual class of Xena ValkyrieVE Tester.
    It is essentially an extended :class:`BaseTester`.


    :param host: tester's address/hostname
    :type host: str
    :param username: username of the user
    :type username: str
    :param password: login password of the tester, defaults to "xena"
    :type password: str, optional
    :param port: the port number for connection establishment, default to 22606
    :type port: int, optional
    :param debug: `True` if debug log output from the tester is needed, and `False` otherwise
    :type debug: int, optional
    """

    def __init__(self, host: str, username: str, password: str = "xena", port: int = 22606, *, debug: bool = False) -> None:
        super().__init__(host=host, username=username, password=password, port=port, debug=debug)

        self._local_states = testers_state.TesterLocalState(host, port)

        self.multiuser = C_MULTIUSER(self._conn)
        """
        Enable or disable the ability to control one resource from several different TCP connections.
        
        :type:  C_MULTIUSER
        """

        self.traffic = C_TRAFFIC(self._conn)
        """
        Starts or stops the traffic on a number of ports on the chassis simultaneously.
        The ports are identified by pairs of integers (module port).

        :type: C_TRAFFIC
        """

        self.traffic_sync = C_TRAFFICSYNC(self._conn)
        """
        This can be used to start traffic simultaneously on multiple chassis. The ports are identified by pairs of integers (module port).

        :type: C_TRAFFICSYNC
        """

        self.version_no_minor = C_VERSIONNO_MINOR(self._conn)
        """
        Get the minor version number of the tester firmware.

        :type: C_VERSIONNO_MINOR
        """

        self.modules: TypeL23Manager = ModulesManager(self._conn, get_module_type)
        """
        Module Index Manager of the L23 VE tester.

        :type: ModulesManager
        """

    @property
    def info(self) -> testers_state.TesterLocalState:
        """Return tester's local state

        :return: tester's local state
        :rtype: GenuineTesterLocalState
        """
        return self._local_states

    async def _setup(self):
        await super()._setup()
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)

        ft_pc = await C_PORTCOUNTS(self._conn).get()
        port_counts = ft_pc.port_counts
        await self.modules.fill_l23(port_counts)
        return self



from typing import NoReturn, TypeVar
from xoa_driver.internals.core.commands.enums import FilterType

T = TypeVar("T")


def prevent_set(inst: T, filter_type: FilterType = FilterType.WORKING) -> T:
    METHOD_NAME = "set"

    def _set(*args, **kwargs) -> NoReturn:
        raise Exception(f"Method {METHOD_NAME} is not allowed")

    if filter_type is FilterType.WORKING and hasattr(inst, METHOD_NAME):
        setattr(inst, METHOD_NAME, _set)
    return inst



from typing import (
    Dict,
    Type,
    Union,
    TYPE_CHECKING
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands.enums import (
    FilterType,
    FilterMode,
)
from . import general


class FilterDefinitionWorking:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self._conn = conn
        self._module_id = module_id
        self._port_id = port_id
        self._flow_index = flow_index
        self._filter_type = FilterType.WORKING

    async def get_mode(self) -> Union[general.ModeBasic, general.ModeExtended]:
        mode_types: Dict[FilterMode, Type[Union[general.ModeBasic, general.ModeExtended]]] = {
            FilterMode.BASIC: general.ModeBasic,
            FilterMode.EXTENDED: general.ModeExtended,
        }
        return await general.get_mode(
            self._conn,
            self._module_id,
            self._port_id,
            self._flow_index,
            self._filter_type,
            mode_types
        )



import struct
from typing import List, cast

from . import constants as const
from .fields.data_types import XmpEmpty
from . import utils
from .struct_header import ProtocolHeader


class Request:
    def __init__(
        self,
        class_name: str,
        indices: List[int],
        cmd_type: int,
        cmd_code: int,
        module_index: int = const.NOTHING,
        port_index: int = const.NOTHING,
        values: object = None,
    ) -> None:
        self.class_name = class_name
        self.header = ProtocolHeader(
            magic_word=const.MAGIC_WORD,
            number_of_indices=len(indices),
            number_of_value_bytes=self._get_values_length(values),
            cmd_code=cmd_code,
            cmd_type=cmd_type,
            module_index=module_index,
            port_index=port_index,
            request_identifier=0,
        )
        self.index_values = indices
        self.values = values
        self.padding = (
            4 - (self.header.number_of_value_bytes % 4)
            if self.header.number_of_value_bytes % 4
            else 0
        )

    def __str__(self) -> str:
        return utils.format_str(
            cast(utils.XmProtocol, self),
            f"padding              : {self.padding}",
        )

    def __repr__(self) -> str:
        return utils.format_repr(cast(utils.XmProtocol, self))

    def __bytes__(self) -> bytes:
        cmd_all = b"".join(
            (
                bytes(self.header),  # type: ignore
                struct.pack(f"!{self.header.number_of_indices}I", *self.index_values),
                self._get_values_bytes(self.values),
                bytes(self.padding),
            )
        )
        return cmd_all

    @staticmethod
    def _get_values_length(values: object) -> int:
        if values is None or not hasattr(values, "__annotations__"):
            return 0
        return sum(
            getattr(values, k, XmpEmpty()).byte_length()
            for k in values.__annotations__.keys()
        )

    @staticmethod
    def _get_values_bytes(values: object) -> bytes:
        if values is None or not hasattr(values, "__annotations__"):
            return b""
        return b"".join(
            bytes(getattr(values, k))
            for k in values.__annotations__.keys()
        )




import functools
from typing import (
    TYPE_CHECKING,
    TypeVar,
    Generic
)
from abc import (
    ABC,
    abstractmethod,
)
from typing_extensions import Self
from xoa_driver.internals.core.commands import enums
from xoa_driver.internals.core.commands import (
    M_MODEL,
    M_PORTCOUNT,
    M_RESERVATION,
    M_RESERVEDBY,
    M_SERIALNO,
    M_VERSIONNO
)
from xoa_driver.internals.utils import attributes as utils

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from . import __interfaces as m_itf
    from xoa_driver.internals.state_storage import modules_state


T = TypeVar('T', bound="modules_state.ModuleLocalState")


class BaseModule(ABC, Generic[T]):
    """Basic Module class.
    """

    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        self._conn = conn
        self.module_id = init_data.module_id
        self.ports_count = init_data.ports_count
        self.reservation = M_RESERVATION(self._conn, self.module_id)
        """Test module's reservation action.
                
        :type: M_RESERVATION
        """

        self.reserved_by = M_RESERVEDBY(self._conn, self.module_id)
        """Test module's reservation status.
                
        :type: M_RESERVEDBY
        """

        self.model = M_MODEL(self._conn, self.module_id)
        """Test module's model.
                
        :type: M_MODEL
        """

        self.serial_number = M_SERIALNO(self._conn, self.module_id)
        """Test module's serial number.
                
        :type: M_SERIALNO
        """

        self.version_number = M_VERSIONNO(self._conn, self.module_id)
        """Test module's version number.
                
        :type: M_VERSIONNO
        """

        self.port_count = M_PORTCOUNT(self._conn, self.module_id)
        """Max port count of the test module.
                
        :type: M_PORTCOUNT
        """

    @property
    @abstractmethod
    def info(self) -> T:
        """
        Module local info.
        """
        raise NotImplementedError()

    def __await__(self):
        return self._setup().__await__()

    @abstractmethod
    async def _setup(self) -> Self:
        raise NotImplementedError()

    def _check_identity(self, request) -> bool:
        return self.module_id == request.header.module_index

    def __is_reservation(self, reserved_status: enums.ReservedStatus) -> bool:
        return self.info.reservation == reserved_status

    is_released = functools.partialmethod(__is_reservation, enums.ReservedStatus.RELEASED)
    """
    Validate if the module is released.
    """

    is_reserved_by_me = functools.partialmethod(__is_reservation, enums.ReservedStatus.RESERVED_BY_YOU)
    """
    Validate if the module is reserved by me.
    """

    on_reservation_change = functools.partialmethod(utils.on_event, M_RESERVATION)
    """
    Register a callback to the event that the module's reservation status changes.
    """

    on_reserved_by_change = functools.partialmethod(utils.on_event, M_RESERVEDBY)
    """
    Register a callback to the event that the module's ownership status changes.
    """

    on_model_change = functools.partialmethod(utils.on_event, M_MODEL)
    """
    Register a callback to the event that the module's legacy model changes.
    """

    on_serial_number_change = functools.partialmethod(utils.on_event, M_SERIALNO)
    """
    Register a callback to the event that the module's serial number status changes.
    """

    on_version_number_change = functools.partialmethod(utils.on_event, M_VERSIONNO)
    """
    Register a callback to the event that the module's version number status changes.
    """



#: All available tester types.

from xoa_driver.internals.hli_v2.testers.l23_tester import L23Tester
from xoa_driver.internals.hli_v2.testers.l23ve_tester import L23VeTester
from xoa_driver.internals.hli_v2.testers.l47_tester import L47Tester
from xoa_driver.internals.hli_v2.testers.l47ve_tester import L47VeTester

import typing

GenericAnyTester = typing.Union[
    "L23Tester",
    "L23VeTester",
    "L47Tester",
    "L47VeTester",
]


__all__ = (
    "L23Tester",
    "L23VeTester",
    "L47Tester",
    "L47VeTester",
    "GenericAnyTester",
)



import setuptools


def main():
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()

    setuptools.setup(
        name="xoa-driver",
        description=(
            "Xena OpenAutomation (XOA) Python API is a driver providing user-friendly communication"
            " interfaces to Xena's physical and virtual Traffic Generation and Analysis (TGA) testers."
            " It provides a rich collection of programming interfaces that can be used to either write"
            " test scripts or develop applications."
        ),
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="Artem Constantinov, Ron Ding, Leonard Yu",
        author_email="aco@xenanetworks.com, rdi@xenanetworks.com, hyu@xenanetworks.com",
        maintainer="Xena Networks",
        maintainer_email="support@xenanetworks.com",
        url="https://github.com/xenanetworks/open-automation-python-api",
        packages=setuptools.find_packages(),
        license='Apache 2.0',
        install_requires=[
        "typing_extensions", 
        "loguru"
        ],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
        ],
        python_requires=">=3.8",
    )


if __name__ == '__main__':
    main()



import sys


if "xoa_driver.internals.hli_v1" in sys.modules:
    raise ImportError("\33[101mUsing xoa_driver and xoa_driver.v2 at the same time is not allowed.\33[0m")

from xoa_driver.internals import warn

warn.resource(
    "xoa_driver.v2 is under development and it subject to changes without notice.",
)



import typing
import functools
from xoa_driver import ports
from xoa_driver.internals.hli_v1 import revisions
from xoa_driver.internals.core.commands import P_CAPABILITIES
from xoa_driver.internals.utils import ports_manager as pm
from xoa_driver.internals.utils.cap_id import CapID

if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from .. import __interfaces as m_itf

from .module_l23_base import ModuleL23


D_FAMELY_ID = CapID(1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
F_FAMELY_ID = CapID(0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)


CombiTypes = typing.Union[
    ports.POdin1G4S4PCombi,
    ports.POdin10G4S2PCombi,
]

CombiTypesB = typing.Union[
    ports.POdin1G4S4PCombi_b,
    ports.POdin10G4S2PCombi_b,
]


async def _port_resolver(conn: "itf.IConnection", module_id: int, port_id: int, port_map: typing.Dict[int, typing.Type]) -> typing.Coroutine[typing.Any, typing.Any, CombiTypes]:
    cap = await P_CAPABILITIES(conn, module_id, port_id).get()
    current_port_id = CapID.create_from_capabilities(cap)
    port_type = port_map[current_port_id.to_int()]
    return await port_type(conn, module_id, port_id)


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-4S-2P-Combi")
class MOdin10G4S2PCombi(ModuleL23):
    """Test module Odin-10G-4S-2P-Combi
    """
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        PORTS_MAP = {
            D_FAMELY_ID.to_int(): ports.POdin1G4S4PCombi,
            F_FAMELY_ID.to_int(): ports.POdin10G4S2PCombi,
        }
        self.ports: pm.PortsCombiManager[CombiTypes] = pm.PortsCombiManager(
            conn=conn,
            resolver=functools.partial(_port_resolver, port_map=PORTS_MAP),
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-4S-2P-Combi
        
        :type: PortsCombiManager
        """


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-4S-2P-Combi[b]")
class MOdin10G4S2PCombi_b(ModuleL23):
    """Test module Odin-10G-4S-2P-Combi[b]
    """
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        PORTS_MAP = {
            D_FAMELY_ID.to_int(): ports.POdin1G4S4PCombi_b,
            F_FAMELY_ID.to_int(): ports.POdin10G4S2PCombi_b,
        }
        self.ports: pm.PortsCombiManager[CombiTypesB] = pm.PortsCombiManager(
            conn=conn,
            resolver=functools.partial(_port_resolver, port_map=PORTS_MAP),
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-4S-2P-Combi[b]
        
        :type: PortsCombiManager
        """



import struct
from typing import (
    Any,
    List,
    Optional,
    get_args,  # py3.8 >
    cast
)
from . import constants as const
from . import utils
from .struct_header import ResponseHeader
from .fields.data_types import (
    XmpDefaultList,
    XmpStr,
)


class NotEnoughBytesLeft(RuntimeError):
    def __init__(self, cmd_name: str, back, curent_field_name: str) -> None:
        self.cmd_name = cmd_name
        all_fields = list(back.__annotations__.keys())
        self.fields = all_fields[all_fields.index(curent_field_name) :]
        self.msg = f"When parsing {self.cmd_name}, there are still fields {self.fields} to be parsed while there is no more bytes!"
        super().__init__(self.msg)


class Response:
    """
    Response serializer
    """

    __slots__ = (
        "class_name",
        "header",
        "raw_data",
        "__cursor",
        "index_values",
        "values",
    )

    def __init__(
        self, header: ResponseHeader, class_name: str, data: bytes, back: Optional[type]
    ) -> None:
        self.class_name = class_name
        self.header = header
        self.raw_data = data
        self.__cursor = 0
        self.index_values: List[int] = self.__parse_indices()
        self.values: Any = None
        if self.header.cmd_type == const.CommandType.COMMAND_VALUE:
            self.values = self.__parse_values(back)
        # elif self.header.cmd_type == CommandType.COMMAND_STATUS:

        del self.__cursor

    def __str__(self) -> str:
        return utils.format_str(
            self, b_str=bytes(self.header) + self.raw_data  # type: ignore
        )

    def __repr__(self) -> str:
        return utils.format_repr(cast(utils.XmProtocol, self))

    def __bytes__(self) -> bytes:
        return bytes(self.header) + self.raw_data

    @property
    def command_status(self) -> Optional[const.CommandStatus]:
        if self.header.cmd_type == const.CommandType.COMMAND_STATUS:
            return const.CommandStatus(self.header.cmd_code)
        return None

    def __parse_indices(self) -> List[int]:
        format = f"!{self.header.number_of_indices}I"
        indices = struct.unpack_from(format, self.raw_data, 0)
        self.__cursor += struct.calcsize(format)
        return list(indices)

    def __parse_values(self, back: Optional[type]):
        if not back:
            return None
        dic = {}
        for field_name, field_type in back.__annotations__.items():
            generic_type = get_args(field_type)[0]
            if generic_type == XmpStr:
                # NOTICE: Here can be issue if we having a command with string and list at the same time
                dic[field_name] = self.__parse_xmp_str(self.__calc_str_len(back))
            else:
                length = self.__calc_xmp_type_length(generic_type)
                value_bytes = self.raw_data[self.__cursor:self.__cursor + length]
                if (not issubclass(generic_type, XmpDefaultList)) and not value_bytes:
                    raise NotEnoughBytesLeft(self.class_name, back, field_name)
                dic[field_name] = generic_type.from_bytes(value_bytes)
                self.__cursor += length
        return back(**dic)

    def __calc_xmp_type_length(self, generic_type) -> int:
        if not issubclass(generic_type, XmpDefaultList):
            return int(generic_type.size)
        else:

            if generic_type.fix_length:
                return int(generic_type.fix_length * generic_type.element_type.size)
            elif generic_type.stop_to_keep:
                if self.header.number_of_value_bytes % 4:
                    real_value_num = self.header.number_of_value_bytes + (
                        4 - self.header.number_of_value_bytes % 4
                    )
                else:
                    real_value_num = self.header.number_of_value_bytes
                # weird parsing with helper types
                return int(real_value_num - generic_type.stop_to_keep)
            return int(self.header.number_of_value_bytes)

    def __calc_str_len(self, back) -> int:
        types_aligned = [
            get_args(field_type)[0] for field_type in back.__annotations__.values()
        ]
        sum_of_fixed_sizes = sum(
            int(generic_type.size) for generic_type in types_aligned
        )
        return self.header.body_size - sum_of_fixed_sizes

    def __parse_xmp_str(self, string_lenght: int) -> bytes:
        raw_data: bytes = self.raw_data[self.__cursor :]
        zero: bytes = b"\x00"
        nvb: int = self.header.number_of_value_bytes
        if zero in raw_data:
            zero_index = raw_data.index(zero)
            length = zero_index if zero_index + self.__cursor <= nvb else nvb
        else:
            length = nvb
        self.__cursor += string_lenght
        return raw_data[:length]

    @property
    def is_ok(self) -> bool:
        return self.get_return_ok() or self.set_return_ok()

    def set_return_ok(self) -> bool:
        return all(
            (
                self.header.cmd_type == const.CommandType.COMMAND_STATUS,
                self.header.cmd_code == const.CommandStatus.OK,
            )
        )

    def get_return_ok(self) -> bool:
        return all(
            (
                self.header.cmd_type == const.CommandType.COMMAND_VALUE,
                self.values is not None
            )
        )



#: L47 Port Packet Engine Commands

from dataclasses import dataclass
import typing

from ..protocol.command_builders import (
    build_get_request,
    build_set_request
)
from .. import interfaces
from ..transporter.token import Token
from ..protocol.fields import data_types as xt
from ..protocol.fields.field import XmpField
from ..registry import register_command
# from .enums import *  # noqa: F403


@register_command
@dataclass
class P4E_ASSIGN:
    """
    Advanced mode only: Assign previously reserved PEs to a port.
    """

    code: typing.ClassVar[int] = 675
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    @dataclass(frozen=True)
    class SetDataAttr:
        mask: XmpField[xt.XmpHex8] = XmpField(xt.XmpHex8)  # eight hex bytes, a bitmask specifying which PEs should be assigned to this port

    @dataclass(frozen=True)
    class GetDataAttr:
        mask: XmpField[xt.XmpHex8] = XmpField(xt.XmpHex8)  # eight hex bytes, a bitmask specifying which PEs should be assigned to this port

    def get(self) -> "Token[GetDataAttr]":
        """Get the assigned PEs on the port.

        :return: previously reserved PEs
        :rtype: P4E_ASSIGN.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, mask: str) -> "Token":
        """Assign the previously reserved PEs to the port.

        :param mask: a bitmask specifying which PEs should be assigned to this port
        :type mask: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, mask=mask))


@register_command
@dataclass
class P4E_AVAILABLE:
    """
    Simple mode only: Report the number of PEs available for allocation.
    """

    code: typing.ClassVar[int] = 676
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    @dataclass(frozen=True)
    class GetDataAttr:
        available_pe_count: XmpField[xt.XmpByte] = XmpField(
            xt.XmpByte
        )  # byte, total number of PEs that can be allocated to the port - including the PEs already allocated to the port.

    def get(self) -> "Token[GetDataAttr]":
        """Get the number of PEs available for allocation.

        :return: the number of PEs available for allocation.
        :rtype: P4E_AVAILABLE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4E_ALLOCATE:
    """
    Simple mode only: Allocate a number of PEs to this port.
    """

    code: typing.ClassVar[int] = 677
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    @dataclass(frozen=True)
    class SetDataAttr:
        pe_count_alloc: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)  # byte, the total number of PEs to allocate to this port - including the PEs already allocated to the port.

    @dataclass(frozen=True)
    class GetDataAttr:
        pe_count_alloc: XmpField[xt.XmpByte] = XmpField(xt.XmpByte)  # byte, the total number of PEs to allocate to this port - including the PEs already allocated to the port.

    def get(self) -> "Token[GetDataAttr]":
        """Get the number of PEs allocated to this port.

        :return: the total number of PEs to allocate to this port - including the PEs already allocated to the port.
        :rtype: P4E_ALLOCATE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, pe_count_alloc: int) -> "Token":
        """Allocate a number of PEs to this port.

        :param pe_count_alloc: the total number of PEs to allocate to this port - including the PEs already allocated to the port.
        :type pe_count_alloc: int
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, pe_count_alloc=pe_count_alloc))


@register_command
@dataclass
class P4E_ALLOCATION_INFO:
    """
    Display information about which PEs that are available for allocation/assignment
    and which are currently allocated/assigned to this port.
    """

    code: typing.ClassVar[int] = 678
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    @dataclass(frozen=True)
    class GetDataAttr:
        available: XmpField[xt.XmpHex8] = XmpField(xt.XmpHex8)  # eight hex bytes (64-bit) mask of available PEs
        allocated: XmpField[xt.XmpHex8] = XmpField(xt.XmpHex8)  # eight hex bytes (64-bit) mask of PEs assigned to this port

    def get(self) -> "Token[GetDataAttr]":
        """Get PEs allocation information.

        :return: PEs allocation information
        :rtype: P4E_ALLOCATION_INFO.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))



import typing
import asyncio
import functools
from typing import TYPE_CHECKING
from typing_extensions import Self
from xoa_driver.internals.core.commands import (
    M_STATUS,
    M_UPGRADE,
    M_UPGRADEPROGRESS,
    M_CFPTYPE,
    M_CFPCONFIGEXT,
    M_CFPCONFIG,
    M_COMMENT,
    M_CAPABILITIES,
    M_CLOCKPPB,
    M_TXCLOCKSOURCE_NEW,
    M_TXCLOCKSTATUS_NEW,
    M_EMULBYPASS,
    M_LATENCYMODE,
)

from xoa_driver.internals.hli_v2 import revisions
from xoa_driver.internals.utils import ports_manager as pm
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.state_storage import modules_state
from xoa_driver.v2 import ports
from . import base_module as bm
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from . import __interfaces as m_itf


class ChTXClock:
    """
    Advanced timing feature (Chimera).
    """
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.source = M_TXCLOCKSOURCE_NEW(conn, module_id)
        """
        The source that drives the TX clock rate of the ports on the test module.
        Representation of M_TXCLOCKSOURCE_NEW
        """
        self.status = M_TXCLOCKSTATUS_NEW(conn, module_id)
        """
        TX clock status of the test module.
        Representation of M_TXCLOCKSTATUS_NEW
        """


class ChCFP:
    """
    CFP test module (Chimera).
    """
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.type = M_CFPTYPE(conn, module_id)
        """
        The transceiver's CFP type currently inserted.
        Representation of M_CFPTYPE
        """
        self.config = M_CFPCONFIG(conn, module_id)
        """
        The CFP configuration of the test module.
        Representation of M_CFPCONFIG
        """
        self.config_extended = M_CFPCONFIGEXT(conn, module_id)
        """
        The CFP configuration of the test module.
        Representation of M_CFPCONFIGEXT
        """


class ChUpgrade:
    """
    Upgrade test module (Chimera).
    """
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.start = M_UPGRADE(conn, module_id)
        """
        Start the upgrade progress of the test module.
        Representation of M_UPGRADE
        """
        self.progress = M_UPGRADEPROGRESS(conn, module_id)
        """
        Upgrade progress status of the test module.
        Representation of M_UPGRADEPROGRESS
        """


class ChTiming:
    """Test module timing and clock configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:

        self.clock_local_adjust = M_CLOCKPPB(conn, module_id)
        """Time adjustment controlling of the local clock of the test module, which drives the TX rate of the test ports.
        Representation of M_CLOCKPPB
        """


class ChAdvancedTiming:
    """Advanced Timing config and control"""
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.clock = ChTXClock(conn, module_id)
        """Advanced timing clock config and status
        """


class ModuleChimera(bm.BaseModule["modules_state.ModuleLocalState"]):
    """
    Representation of a Chimera module on physical tester.
    """
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)

        self._local_states = modules_state.ModuleLocalState()

        self.tx_clock = ChTXClock(conn, self.module_id)
        """
        Advanced timing feature (Chimera).
        """

        self.cfp = ChCFP(conn, self.module_id)
        """
        CFP test module (Chimera).
        """

        self.upgrade = ChUpgrade(conn, self.module_id)
        """
        Upgrade test module (Chimera).
        """

        self.capabilities = M_CAPABILITIES(conn, self.module_id)
        """
        Test module's capabilities.
        Representation of M_CAPABILITIES
        """

        self.comment = M_COMMENT(conn, self.module_id)
        """
        Test module's description.
        Representation of M_COMMENT
        """

        self.status = M_STATUS(conn, self.module_id)
        """
        Test module's status.
        Representation of M_STATUS
        """

        self.clock_ppb = M_CLOCKPPB(conn, self.module_id)
        """
        Test module's local clock adjustment.
        Representation of M_CLOCKPPB
        """

        self.emulator_bypass_mode = M_EMULBYPASS(conn, self.module_id)
        """
        Bypass mode of the Chimera module.
        Representation of M_EMULBYPASS
        """

        self.latency_mode = M_LATENCYMODE(conn, self.module_id)
        """
        Latency mode of the Chimera module.
        Representation of M_LATENCYMODE
        """

        self.timing = ChTiming(conn, self.module_id)
        """Test module's timing configuration."""

        self.bypass_mode = M_EMULBYPASS(conn, self.module_id)
        """
        Bypass mode of the Chimera module.
        Representation of M_EMULBYPASS
        """

        self.adv_timing = ChAdvancedTiming(conn, self.module_id)

        self.ports: pm.PortsManager["ports.PortChimera"] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PortChimera,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """
        Port index manager of the Chimera module.
        """

    @property
    def info(self) -> modules_state.ModuleLocalState:
        return self._local_states

    async def _setup(self) -> Self:
        await asyncio.gather(
            self._local_states.initiate(self),
            self.ports.fill()
        )
        self._local_states.register_subscriptions(self)
        return self

    on_cfp_type_change = functools.partialmethod(utils.on_event, M_CFPTYPE)
    """
    Register a callback to the event that the module's CFP type changes.
    """

    on_cfp_config_change = functools.partialmethod(utils.on_event, M_CFPCONFIG)
    """
    Register a callback to the event that the module's CFP configuration changes.
    """

    on_cfp_config_extended_change = functools.partialmethod(utils.on_event, M_CFPCONFIGEXT)
    """
    Register a callback to the event that the module's CFP configuration changes.
    """

    on_status_change = functools.partialmethod(utils.on_event, M_STATUS)
    """
    Register a callback to the event that the module's model changes.
    """

    on_latency_mode_change = functools.partialmethod(utils.on_event, M_LATENCYMODE)
    """
    Register a callback to the event that the module's latency mode changes.
    """


@typing.final
@revisions.register_chimera_module(rev="Chimera-100G-5S-2P")
class MChi100G5S2P(ModuleChimera):
    """Chimera module Chi-100G-5S-2P"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PChi100G5S2P] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PChi100G5S2P,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Chi-100G-5S-2P"""


@typing.final
@revisions.register_chimera_module(rev="Chimera-100G-5S-2P[b]")
class MChi100G5S2P_b(ModuleChimera):
    """Chimera module Chi-100G-5S-2P[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PChi100G5S2P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PChi100G5S2P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Chi-100G-5S-2P[b]"""


@typing.final
@revisions.register_chimera_module(rev="Chimera-40G-2S-2P")
class MChi40G2S2P(ModuleChimera):
    """Chimera module Chi-40G-2S-2P"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PChi40G2S2P] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PChi40G2S2P,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Chi-40G-2S-2P"""



import asyncio
import typing
from typing_extensions import Self
from xoa_driver.internals.core.commands import (
    M_COMMENT,
    M_CAPABILITIES,
    M_MULTIUSER
)
from xoa_driver import ports
from xoa_driver.internals.utils import ports_manager as pm
from xoa_driver.internals.state_storage import modules_state
from . import base_module as bm
if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from . import __interfaces as m_itf


class ModuleL23VE(bm.BaseModule["modules_state.ModuleLocalState"]):
    """
    This is a conceptual class of L23 test module on ValkyrieVE tester.
    """

    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)

        self._local_states = modules_state.ModuleLocalState()

        self.comment = M_COMMENT(conn, self.module_id)
        """Test module's description.
        
        :type: M_COMMENT
        """

        self.capabilities = M_CAPABILITIES(conn, self.module_id)
        """Test module's capabilities.
        
        :type: M_CAPABILITIES
        """

        self.multiuser = M_MULTIUSER(conn, self.module_id)
        """If multiple users are allowed to control the same test module.
        
        :type: M_MULTIUSER
        """

        self.ports: pm.PortsManager[ports.PortL23VE] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PortL23VE,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """L23 VE Port index manager of this test module.
        
        :type: PortsManager
        """

    @property
    def info(self) -> modules_state.ModuleLocalState:
        """Return the module's local state

        :return: the module's local state
        :rtype: ModuleLocalState
        """

        return self._local_states

    async def _setup(self) -> Self:
        await asyncio.gather(
            self._local_states.initiate(self),
            self.ports.fill()
        )
        self._local_states.register_subscriptions(self)
        return self



from typing import (
    TYPE_CHECKING,
    Tuple,
)
from dataclasses import astuple
from xoa_driver.internals.core.commands.enums import ImpairmentTypeIndex
from xoa_driver.internals.core.commands.ped_commands import PED_ENABLE, PED_ONESHOTSTATUS, PED_SCHEDULE

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from xoa_driver.internals.core.commands import (
    PE_FCSDROP,
    PE_TPLDMODE,
    PE_COMMENT,
    PE_INDICES,
    PE_LATENCYRANGE,
    PE_CORRUPT,
    PE_MISORDER,
    PE_BANDPOLICER,
    PE_BANDSHAPER,
    PE_DROPTOTAL,
    PE_LATENCYTOTAL,
    PE_DUPTOTAL,
    PE_MISTOTAL,
    PE_CORTOTAL,
    PE_JITTERTOTAL,
    PE_CLEAR,
    PE_FLOWDROPTOTAL,
    PE_FLOWLATENCYTOTAL,
    PE_FLOWDUPTOTAL,
    PE_FLOWMISTOTAL,
    PE_FLOWCORTOTAL,
    PE_FLOWJITTERTOTAL,
    PE_FLOWCLEAR,
)

from .reception_statistics import ReceptionStatistics
from .transmission_statistics import TransmissionStatistics
from .pe_distribution import ImpairmentDistribution
from .filter_definition import (
    shadow,
    working,
)


class CTotalFlow:
    """Total flow statistics."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.dropped = PE_FLOWDROPTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets dropped in a flow.
        Representation of PE_FLOWDROPTOTAL
        """

        self.delayed = PE_FLOWLATENCYTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets delayed in a flow.
        Representation of PE_FLOWLATENCYTOTAL
        """

        self.duplicated = PE_FLOWDUPTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets duplicate in a flow.
        Representation of PE_FLOWDUPTOTAL
        """

        self.misordered = PE_FLOWMISTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets misordered in a flow.
        Representation of PE_FLOWMISTOTAL
        """

        self.corrupted = PE_FLOWCORTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets corrupted in a flow.
        Representation of PE_FLOWCORTOTAL
        """

        self.jittered = PE_FLOWJITTERTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets jittered in a flow.
        Representation of PE_FLOWJITTERTOTAL
        """


class StatisticsTotals:
    """Total port statistics."""

    def __init__(self, conn, module_id, port_id) -> None:
        self.dropped = PE_DROPTOTAL(conn, module_id, port_id)
        """Statistics of all packets dropped on the port.
        Representation of PE_DROPTOTAL
        """

        self.delayed = PE_LATENCYTOTAL(conn, module_id, port_id)
        """Statistics of all packets delayed on the port.
        Representation of PE_LATENCYTOTAL
        """

        self.duplicated = PE_DUPTOTAL(conn, module_id, port_id)
        """Statistics of all packets duplicated on the port.
        Representation of PE_DUPTOTAL
        """

        self.misordered = PE_MISTOTAL(conn, module_id, port_id)
        """Statistics of all packets mirordered on the port.
        Representation of PE_MISTOTAL
        """

        self.corrupted = PE_CORTOTAL(conn, module_id, port_id)
        """Statistics of all packets corrupted on the port.
        Representation of PE_CORTOTAL
        """

        self.jittered = PE_JITTERTOTAL(conn, module_id, port_id)
        """Statistics of all packets jittered on the port.
        Representation of PE_JITTERTOTAL
        """

        self.fcs_errors = PE_FCSDROP(conn, module_id, port_id)
        """Action on FCS errors.
        Representation of PE_FCSDROP
        """

        self.clear = PE_CLEAR(conn, module_id, port_id)
        """Clear impairment statistics.
        Representation of PE_CLEAR
        """


class CFlowStatistics:
    """Per flow statistics."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.rx = ReceptionStatistics(conn, module_id, port_id, flow_index)
        """RX statistics."""

        self.tx = TransmissionStatistics(conn, module_id, port_id, flow_index)
        """TX statistics."""

        self.total = CTotalFlow(conn, module_id, port_id, flow_index)
        """Total flow statistics."""

        self.clear = PE_FLOWCLEAR(conn, module_id, port_id, flow_index)
        """Clear the impairment statistics on a flow.
        Representation of PE_FLOWCLEAR
        """


class CBandwidth:
    """Bandwidth configuration."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.policer = PE_BANDPOLICER(conn, module_id, port_id, flow_index)
        """Bandwidth policer configuration.
        Representation of PE_BANDPOLICER
        """

        self.shaper = PE_BANDSHAPER(conn, module_id, port_id, flow_index)
        """Bandwidth shaper configuration.
        Representation of PE_BANDSHAPER
        """


class ChimeraPE:
    """Port Emulation"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self._conn = conn
        self.module_id = module_id
        self.port_id = port_id
        self.drop_fcs_errors = PE_FCSDROP(conn, module_id, port_id)
        """Action on FCS errors.
        Representation of PE_FCSDROP
        """

        self.clear = PE_CLEAR(conn, module_id, port_id)
        """Clear impairment statistics.
        Representation of PE_CLEAR
        """

        self.statistics = StatisticsTotals(conn, module_id, port_id)
        """Total port statistics."""

        self.tpld_mode = PE_TPLDMODE(conn, module_id, port_id)
        """TPLD mode.
        Representation of PE_TPLDMODE
        """

        self.flow: Tuple["CFlow", ...] = tuple()

    def __await__(self):
        return self._setup().__await__()

    async def _setup(self) -> None:
        indices = astuple(await PE_INDICES(self._conn, self.module_id, self.port_id).get())
        self.flow = tuple(
            CFlow(self._conn, self.module_id, self.port_id, idx)
            for idx in indices
        )


class CTotalFlowStats:
    """Total flow statistics."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.dropped = PE_FLOWDROPTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets dropped in a flow.
        Representation of PE_FLOWDROPTOTAL
        """

        self.delayed = PE_FLOWLATENCYTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets delayed in a flow.
        Representation of PE_FLOWLATENCYTOTAL
        """

        self.duplicated = PE_FLOWDUPTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets duplicate in a flow.
        Representation of PE_FLOWDUPTOTAL
        """

        self.misordered = PE_FLOWMISTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets misordered in a flow.
        Representation of PE_FLOWMISTOTAL
        """

        self.corrupted = PE_FLOWCORTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets corrupted in a flow.
        Representation of PE_FLOWCORTOTAL
        """

        self.jittered = PE_FLOWJITTERTOTAL(conn, module_id, port_id, flow_index)
        """Statistics of all packets jittered in a flow.
        Representation of PE_FLOWJITTERTOTAL
        """


class CPerFlowStats:
    """Per flow statistics."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.rx = ReceptionStatistics(conn, module_id, port_id, flow_index)
        """RX statistics."""

        self.tx = TransmissionStatistics(conn, module_id, port_id, flow_index)
        """TX statistics."""

        self.total = CTotalFlowStats(conn, module_id, port_id, flow_index)
        """Total flow statistics."""

        self.clear = PE_FLOWCLEAR(conn, module_id, port_id, flow_index)
        """Clear the impairment statistics on a flow.
        Representation of PE_FLOWCLEAR
        """


class CDropImpairment:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.distribution = ImpairmentDistribution(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DROP)

        self.schedule = PED_SCHEDULE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DROP)
        """Impairment scheduling configuration.
        Representation of PED_SCHEDULE
        """

        self.one_shot_status = PED_ONESHOTSTATUS(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DROP)
        """One-shot status.
        Representation of PED_ONESHOTSTATUS
        """

        self.enable = PED_ENABLE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DROP)
        """Impairment distribution control.
        Representation of PED_ENABLE
        """


class CCorruptionImpairment:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:

        self.type = PE_CORRUPT(conn, module_id, port_id, flow_index)
        """Bandwidth policer configuration.
        Representation of PE_BANDPOLICER
        """

        self.distribution = ImpairmentDistribution(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.CORRUPTION)

        self.schedule = PED_SCHEDULE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.CORRUPTION)
        """Impairment scheduling configuration.
        Representation of PED_SCHEDULE
        """

        self.one_shot_status = PED_ONESHOTSTATUS(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.CORRUPTION)
        """One-shot status.
        Representation of PED_ONESHOTSTATUS
        """

        self.enable = PED_ENABLE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DROP)
        """Impairment distribution control.
        Representation of PED_ENABLE
        """


class CMisorderingImpairment:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.depth = PE_MISORDER(conn, module_id, port_id, flow_index)
        """Misordering depth
        Representation of PE_MISORDER
        """

        self.distribution = ImpairmentDistribution(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.MISORDER)

        self.schedule = PED_SCHEDULE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.MISORDER)
        """Impairment scheduling configuration.
        Representation of PED_SCHEDULE
        """

        self.one_shot_status = PED_ONESHOTSTATUS(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.MISORDER)
        """One-shot status.
        Representation of PED_ONESHOTSTATUS
        """

        self.enable = PED_ENABLE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DROP)
        """Impairment distribution control.
        Representation of PED_ENABLE
        """


class CLatencyJitterImpairment:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:

        self.range = PE_LATENCYRANGE(conn, module_id, port_id, flow_index)
        """Flow latency range.
        Representation of PE_LATENCYRANGE
        """

        self.distribution = ImpairmentDistribution(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.LATENCYJITTER)

        self.schedule = PED_SCHEDULE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.LATENCYJITTER)
        """Impairment scheduling configuration.
        Representation of PED_SCHEDULE
        """

        self.one_shot_status = PED_ONESHOTSTATUS(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.LATENCYJITTER)
        """One-shot status.
        Representation of PED_ONESHOTSTATUS
        """

        self.enable = PED_ENABLE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DROP)
        """Impairment distribution control.
        Representation of PED_ENABLE
        """


class CDuplicationImpairment:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:

        self.distribution = ImpairmentDistribution(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DUPLICATION)

        self.schedule = PED_SCHEDULE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DUPLICATION)
        """Impairment scheduling configuration.
        Representation of PED_SCHEDULE
        """

        self.one_shot_status = PED_ONESHOTSTATUS(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DUPLICATION)
        """One-shot status.
        Representation of PED_ONESHOTSTATUS
        """

        self.enable = PED_ENABLE(conn, module_id, port_id, flow_index, ImpairmentTypeIndex.DROP)
        """Impairment distribution control.
        Representation of PED_ENABLE
        """


class CShaperImpairment:
    """Bandwidth shaper impairment configuration."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.config = PE_BANDSHAPER(conn, module_id, port_id, flow_index)
        """Bandwidth shaper configuration.
        Representation of PE_BANDSHAPER
        """


class CPolicerImpairment:
    """Bandwidth policer impairment configuration."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.config = PE_BANDPOLICER(conn, module_id, port_id, flow_index)
        """Bandwidth policer configuration.
        Representation of PE_BANDPOLICER
        """


class CFlow:
    """Flow settings."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.comment = PE_COMMENT(conn, module_id, port_id, flow_index)
        """Flow description.
        Representation of PE_COMMENT
        """

        self.statistics = CPerFlowStats(conn, module_id, port_id, flow_index)
        """Flow statistics."""

        self.shadow_filter = shadow.FilterDefinitionShadow(conn, module_id, port_id, flow_index)
        """Shadow copy."""

        self.working_filter = working.FilterDefinitionWorking(conn, module_id, port_id, flow_index)
        """Working copy."""

        self.drop = CDropImpairment(conn, module_id, port_id, flow_index)

        self.misordering = CMisorderingImpairment(conn, module_id, port_id, flow_index)

        self.latency_jitter = CLatencyJitterImpairment(conn, module_id, port_id, flow_index)

        self.duplication = CDuplicationImpairment(conn, module_id, port_id, flow_index)

        self.corruption = CCorruptionImpairment(conn, module_id, port_id, flow_index)

        self.policer = CPolicerImpairment(conn, module_id, port_id, flow_index)

        self.shaper = CShaperImpairment(conn, module_id, port_id, flow_index)


from __future__ import annotations
from typing import (
    NamedTuple,
    Protocol,
    Any,
)
import ctypes as c
from . import constants as const


class XmpHeader(Protocol):
    magic_word: c.c_char_p
    number_of_indices: c.c_ushort
    number_of_value_bytes: c.c_ushort
    command_parameter: c.c_ushort
    module_index: c.c_ubyte
    port_index: c.c_ubyte
    request_identifier: c.c_uint32

    @property
    def cmd_type(self) -> int: ...  # noqa: E704

    @property
    def cmd_code(self) -> int: ...   # noqa: E704

    @property
    def body_size(self) -> int: ...   # noqa: E704


class XmProtocol(Protocol):
    header: XmpHeader
    class_name: str
    index_values: list[int]
    values: Any

    def __bytes__(self) -> bytes: ...  # noqa: E704


class CodeTypeStr(NamedTuple):
    type: str  # name of command type
    code: str  # name of command status code, or command name


def repr_bytes(data: bytes) -> list[str]:
    return data.hex(",").split(",")


def get_code_str(x: XmProtocol) -> CodeTypeStr:
    code, ty = (
        x.header.cmd_code,
        x.header.cmd_type,
    )
    ty_str = const.CommandType(ty).name
    if ty == const.CommandType.COMMAND_STATUS:
        code_str = const.CommandStatus(code).name
    else:
        code_str = x.class_name
    return CodeTypeStr(ty_str, code_str)


def format_repr(obj: XmProtocol) -> str:
    ty_str, code_str = get_code_str(obj)
    return (
        f"{str(obj.header.module_index):3s} "
        f"{str(obj.header.port_index):3s} "
        f"{str(obj.index_values):10s} "
        f"{str(obj.header.request_identifier):5s} "
        f"{str(obj.class_name):25s} "
        f"{str(code_str):25s} "
        f"{str(ty_str):10s} "
        f"{obj.values}"
    )


def format_str(obj: XmProtocol, *args: str, b_str: bytes | None = None) -> str:
    bin_str = repr_bytes(bytes(obj) if not b_str else b_str)
    # The Response object are not having __bytes__ method,
    # but this function explicitly use <b_str> for <Response> and no other use cases forthis util so it's dosent mater
    (ty_str, code_str) = get_code_str(obj)
    obj_name = type(obj).__name__
    if obj_name == 'Response':
        cmd_p = 'Replied' if obj.header.request_identifier != 0 else 'Pushed'
    else:
        cmd_p = [code_str, ty_str]

    return "\n" + "\n".join(
        (
            f"{obj_name}              : {bin_str}",
            f"class_name           : {obj.class_name}",
            f"magic_word           : {obj.header.magic_word}",
            f"number_of_indices    : {obj.header.number_of_indices}",
            f"number_of_value_bytes: {obj.header.number_of_value_bytes}",
            f"command_parameter    : {obj.header.command_parameter}:{cmd_p}",
            f"module_index         : {obj.header.module_index}",
            f"port_index           : {obj.header.port_index}",
            f"request_identifier   : {obj.header.request_identifier}",
            f"index_values         : {obj.index_values}",
            f"values               : {obj.values}",
        ) + args
    )
